
from nipype.interfaces.ants.base import (ANTSCommand, ANTSCommandInputSpec)
from nipype.interfaces.base import TraitedSpec, File, traits, isdefined


class DenoiseImageInputSpec(ANTSCommandInputSpec):
    dimension = traits.Enum(
        2,
        3,
        4,
        argstr="-d %d",
        desc="This option forces the image to be treated "
        "as a specified-dimensional image. If not "
        "specified, the program tries to infer the "
        "dimensionality from the input image.",
    )
    input_image = File(
        exists=True,
        argstr="-i %s",
        mandatory=True,
        desc="A scalar image is expected as input for noise correction.",
    )

    mask_image = File(
        exists=True,
        argstr="-x %s",
        mandatory=False,
        desc="Restriction of computation on the mask only",
    )
    noise_model = traits.Enum(
        "Gaussian",
        "Rician",
        argstr="-n %s",
        usedefault=True,
        desc=("Employ a Rician or Gaussian noise model."),
    )
    shrink_factor = traits.Int(
        default_value=1,
        usedefault=True,
        argstr="-s %s",
        desc=(
            "Running noise correction on large images can"
            " be time consuming. To lessen computation time,"
            " the input image can be resampled. The shrink"
            " factor, specified as a single integer, describes"
            " this resampling. Shrink factor = 1 is the default."
        ),
    )
    output_image = File(
        argstr="-o %s",
        name_source=["input_image"],
        hash_files=False,
        keep_extension=True,
        name_template="%s_noise_corrected",
        desc="The output consists of the noise corrected"
        " version of the input image.",
    )
    save_noise = traits.Bool(
        False,
        mandatory=True,
        usedefault=True,
        desc=("True if the estimated noise should be saved to file."),
        xor=["noise_image"],
    )
    noise_image = File(
        name_source=["input_image"],
        hash_files=False,
        keep_extension=True,
        name_template="%s_noise",
        desc="Filename for the estimated noise.",
    )
    verbose = traits.Bool(False, argstr="-v", desc=("Verbose output."))


class DenoiseImageOutputSpec(TraitedSpec):
    output_image = File(exists=True)
    noise_image = File()


class DenoiseImage(ANTSCommand):
    """
    Examples
    --------
    >>> import copy
    >>> from nipype.interfaces.ants import DenoiseImage
    >>> denoise = DenoiseImage()
    >>> denoise.inputs.dimension = 3
    >>> denoise.inputs.input_image = 'im1.nii'
    >>> denoise.cmdline
    'DenoiseImage -d 3 -i im1.nii -n Gaussian -o im1_noise_corrected.nii -s 1'

    >>> denoise_2 = copy.deepcopy(denoise)
    >>> denoise_2.inputs.output_image = 'output_corrected_image.nii.gz'
    >>> denoise_2.inputs.noise_model = 'Rician'
    >>> denoise_2.inputs.shrink_factor = 2
    >>> denoise_2.cmdline
    'DenoiseImage -d 3 -i im1.nii -n Rician
    -o output_corrected_image.nii.gz -s 2'

    >>> denoise_3 = DenoiseImage()
    >>> denoise_3.inputs.input_image = 'im1.nii'
    >>> denoise_3.inputs.save_noise = True
    >>> denoise_3.cmdline
    'DenoiseImage -i im1.nii -n Gaussian
    -o [ im1_noise_corrected.nii, im1_noise.nii ] -s 1'

    """

    input_spec = DenoiseImageInputSpec
    output_spec = DenoiseImageOutputSpec
    _cmd = "DenoiseImage"

    def _format_arg(self, name, trait_spec, value):
        if (name == "output_image") and (
            self.inputs.save_noise or isdefined(self.inputs.noise_image)
        ):
            newval = "[ {}, {} ]".format(
                self._filename_from_source("output_image"),
                self._filename_from_source("noise_image"),
            )
            return trait_spec.argstr % newval

        return super()._format_arg(name, trait_spec, value)



# on the fly function for checking what is passed in "connect"
# should end up in ~ nipype.utils.misc


def show_files(files):
    print(files)
    return files


def print_val(val):
    print(val)
    return val


def print_nii_data(nii_file):
    import nibabel as nib

    data = nib.load(nii_file).get_data()
    print(nii_file, data)
    return nii_file


def get_elem(list_elem, index_elem):
    assert isinstance(list_elem, list), 'Error, list_elem should be a list'
    assert 0 <= index_elem and index_elem < len(list_elem), \
        ('error with index {}, does not match a list with {} elements'.format(
            index_elem, len(list_elem)))

    elem = list_elem[index_elem]
    print(elem)

    return elem


def get_index(list_elem, index_elem):
    assert isinstance(list_elem, list), 'Error, list_elem should be a list'
    assert 1 <= index_elem and index_elem < len(list_elem)+1, \
        ('error with index {}, shoud start at 1 and be smaller than {}'.format(
            index_elem, len(list_elem)+1))

    return list_elem[index_elem-1]


def get_pattern(list_elem, pattern):

    assert isinstance(list_elem, list), 'Error, list_elem should be a list'

    for elem in list_elem:
        if pattern in elem:
            print("Found {} in {}".format(pattern, elem))
            return elem

    assert False, "Could not find {} in {}".format(pattern, elem)


def get_list_length(list_elem):
    assert isinstance(list_elem, list), 'Error, list_elem should be a list'
    return len(list_elem)


def get_first_elem(elem):
    print(elem)
    if isinstance(elem, list):
        print("OK, is list")
        assert len(elem) == 1, "Error, list should contain only one element"
        return elem[0]
    else:
        print("not a list")
        return elem


def gzip(unzipped_file):

    import os
    import shutil
    import subprocess

    head, tail = os.path.split(unzipped_file)

    dest = os.path.abspath(tail)

    print("Copying {} to {}".format(unzipped_file, dest))
    shutil.copy(unzipped_file, dest)
    cmd_line = "gzip {}".format(dest)

    subprocess.check_output(cmd_line, shell=True)

    zipped_file = dest + ".gz"

    assert os.path.exists(zipped_file), \
        "Error, {} should exists".format(zipped_file)

    return zipped_file


def gunzip(zipped_file):
    import subprocess
    import shutil
    import os

    head, tail = os.path.split(zipped_file)

    dest = os.path.abspath(tail)

    shutil.copy(zipped_file, dest)

    if zipped_file[-3:] == ".gz":
        subprocess.check_output("gunzip " + dest, shell=True)
    else:
        ValueError("Non GZip file given")

    unzipped_file = dest[:-3]
    return unzipped_file


def merge_3_elem_to_list(elem1, elem2, elem3):
    return [elem1, elem2, elem3]


def parse_key(params, key):

    from nipype.interfaces.base import isdefined

    def _parse_key(params, cur_key):
        if cur_key in params.keys():
            return params[cur_key]
        else:
            "Error, key {} was not found in {}".format(key, params.keys())
            return {}

    if isdefined(params):
        if isinstance(key, tuple):
            for cur_key in key:
                params = _parse_key(params, cur_key)
        else:
            params = _parse_key(params, key)

        return params

    else:
        return {}


def list_input_files(list_T1, list_T2):

    print("list_T1:", list_T1)
    print("list_T2:", list_T2)

    return ""



# -*- coding: utf-8 -*-
#
# macapype documentation build configuration file, created by
# sphinx-quickstart on Tue Nov 27 18:47:41 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

from datetime import date
import sphinx_gallery  # noqa
from sphinx_gallery.sorting import FileNameSortKey
import sphinx_bootstrap_theme

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.intersphinx',
    'numpydoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx_gallery.gen_gallery',
    'sphinxcontrib.fulltoc'
]

# generate autosummary even if no references
autosummary_generate = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Macapype'
td = date.today()
copyright = u'%s, Macapype Developers (macatools). Last updated on %s' % (td.year,
                                                                td.isoformat())

author = 'Macapype Developers (macatools)'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

import macapype
# The short X.Y version.
version = macapype.__version__

# The full version, including alpha/beta/rc tags.
release = macapype.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'bootstrap'
html_theme_path = sphinx_bootstrap_theme.get_html_theme_path()

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_logo = "_images/logo_macapype_0.3.jpg"

html_sidebars = {'**': ['localtoc.html'],
   'using/windows': ['windowssidebar.html']}

html_theme_options = {
    'navbar_title': 'Macapype',
    'bootswatch_theme': "flatly",
    'navbar_sidebarrel': False,
    'bootstrap_version': "3",
    #'navbar_links': [
        #("Gallery", "auto_examples/index"),
        #("API", "api"),
        #("Tutorial", "tutorial"),
        #("Installation", "install"),
        #("Github", "https://github.com/macatools/macapype", True),
    #]
    'navbar_links': [
        ("Github", "https://github.com/davidmeunier79/macapype", True),
    ]
    }

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
#html_static_path = ['_static']


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'macapypedoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

latex_elements = {
    #'papersize': 'a4paper',
    'papersize': 'letterpaper',
    'pointsize': '12pt',
    'preamble': r'''
    \usepackage[none]{hyphenat}
    \usepackage[document]{ragged2e}
    '''
}


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'macapype.tex', 'macapype Documentation',
     'David Meunier', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'macapype', 'macapype Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'macapype', 'macapype Documentation',
     author, 'macapype', 'One line description of project.',
     'Miscellaneous'),
]

intersphinx_mapping = {"python": ("https://docs.python.org", None)}

sphinx_gallery_conf = {
    'examples_dirs': '../examples_doc',
    'gallery_dirs': 'auto_examples',
    'filename_pattern': '^((?!sgskip).)*$',
    'backreferences_dir': 'generated',
    'within_subsection_order': FileNameSortKey,
    "reference_url": {
    }
}

import sys
import os.path as op

path = op.join(op.dirname(__file__), "../examples_doc")
sys.path.insert(0, path)

#from visbrain.config import CONFIG
#CONFIG['MPL_RENDER'] = True



import os.path as op

from itertools import product

from pathlib import Path

from macapype.utils.utils_tests import (make_tmp_dir, format_template,
                                        load_test_data)
from macapype.pipelines.full_pipelines import \
    (create_full_ants_subpipes, create_full_T1_ants_subpipes,
     create_full_spm_subpipes)

cwd = Path.cwd()
data_path = make_tmp_dir()


def test_create_full_ants_subpipes_no_args():

    params = {
        "general":
        {
            "template_name": "NMT_v1.3better"
        },
        "short_preparation_pipe":
        {
        },
        "fast":
        {
        },
        "extract_pipe":
        {
        },
        "debias":
        {
        },
        "brain_segment_pipe":
        {
            "reg":
                {},
            "segment_atropos_pipe":
                {}
        }
    }

    # params template
    template_name = params["general"]["template_name"]

    template_dir = load_test_data(template_name, data_path)
    params_template = format_template(template_dir, template_name)

    # running workflow
    segment_pnh = create_full_ants_subpipes(
        params=params, params_template=params_template,
        params_template_stereo=params_template,
        name="test_create_full_ants_subpipes_no_args")

    segment_pnh.base_dir = data_path

    segment_pnh.write_graph(graph2use="colored")
    assert op.exists(
        op.join(data_path, "test_create_full_ants_subpipes_no_args",
                "graph.png"))


def test_create_full_ants_subpipes_all_default_params():

    import os
    import json

    softs = ["ants"]
    all_species = ["marmo", "macaque", 'baboon']
    spaces = ["native", "template"]
    pads = [True, False]

    for soft, species, space, pad in product(softs, all_species, spaces, pads):

        print("*** Testing soft {} with species {}".format(soft, species))

        wf_dir = (cwd / "workflows").resolve()
        print(wf_dir)

        params_file = os.path.join(wf_dir,  "params_segment_{}_{}.json".format(
            species, soft))

        assert op.exists(params_file), \
            "Could not find params_file {}".format(params_file)

        params = json.load(open(params_file))

        # params template
        template_name = params["general"]["template_name"]

        template_dir = load_test_data(template_name, data_path)
        params_template = format_template(template_dir, template_name)

        # running workflow
        segment_pnh = create_full_ants_subpipes(
            params=params, params_template=params_template,
            params_template_stereo=params_template,
            pad=pad,
            name="test_create_full_ants_subpipes_all_default_params")

        segment_pnh.base_dir = data_path

        segment_pnh.write_graph(graph2use="colored")
        assert op.exists(
            op.join(data_path,
                    "test_create_full_ants_subpipes_all_default_params",
                    "graph.png"))


def test_create_full_ants_t1_subpipes_all_default_params():

    import os
    import json

    softs = ["ants"]
    all_species = ["marmo", "macaque", 'baboon']
    spaces = ["native", "template"]
    pads = [True, False]

    for soft, species, space, pad in product(softs, all_species, spaces, pads):

        print("*** Testing soft {} with species {}".format(soft, species))

        wf_dir = (cwd / "workflows").resolve()
        print(wf_dir)

        params_file = os.path.join(wf_dir,  "params_segment_{}_{}.json".format(
            species, soft))

        assert op.exists(params_file), \
            "Could not find params_file {}".format(params_file)

        params = json.load(open(params_file))

        # params template
        template_name = params["general"]["template_name"]

        template_dir = load_test_data(template_name, data_path)
        params_template = format_template(template_dir, template_name)

        # running workflow
        segment_pnh = create_full_T1_ants_subpipes(
            params=params, params_template=params_template,
            params_template_stereo=params_template,
            pad=pad,
            name="test_create_full_ants_subpipes_all_default_params")

        segment_pnh.base_dir = data_path

        segment_pnh.write_graph(graph2use="colored")

        assert op.exists(
            op.join(data_path,
                    "test_create_full_ants_subpipes_all_default_params",
                    "graph.png"))


def test_create_full_spm_subpipes_all_default_params():

    import os
    import json

    softs = ["spm"]
    all_species = ["macaque", 'baboon']
    spaces = ["native", "template"]
    pads = [True, False]

    for soft, species, space, pad in product(softs, all_species, spaces, pads):

        print("*** Testing soft {} with species {}".format(soft, species))

        wf_dir = (cwd / "workflows").resolve()
        print(wf_dir)

        params_file = os.path.join(wf_dir,  "params_segment_{}_{}.json".format(
            species, soft))

        assert op.exists(params_file), \
            "Could not find params_file {}".format(params_file)

        params = json.load(open(params_file))

        # params template
        template_name = params["general"]["template_name"]

        template_dir = load_test_data(template_name, data_path)
        params_template = format_template(template_dir, template_name)

        # running workflow
        segment_pnh = create_full_spm_subpipes(
            params=params, params_template=params_template,
            params_template_stereo=params_template, space=space,
            pad=pad,
            name="test_create_full_ants_subpipes_all_default_params")

        segment_pnh.base_dir = data_path

        segment_pnh.write_graph(graph2use="colored")
        assert op.exists(
            op.join(data_path,
                    "test_create_full_ants_subpipes_all_default_params",
                    "graph.png"))


if __name__ == '__main__':

    test_create_full_spm_subpipes_all_default_params()
    test_create_full_ants_subpipes_all_default_params()
    test_create_full_ants_t1_subpipes_all_default_params()




from macapype.utils.utils_bids import create_datasink


def test_create_datasink_sub_ses():
    iterables = [('subject', "mysub"), ('session', 'myses')]
    datasink = create_datasink(iterables)
    print(datasink)

    assert True


def test_create_datasink_sub():
    iterables = [('subject', "mysub")]
    datasink = create_datasink(iterables)
    print(datasink)

    assert True



from nipype.interfaces.base import traits

from macapype.utils.misc import parse_key


def test_parse_key_empty():
    params = {}

    val = parse_key(params, "test")
    assert not val


def test_parse_key_Undefined():
    params = traits.Undefined

    val = parse_key(params, "test")
    assert not val



import pytest
import os.path as op
from macapype.nodes.extract_brain import T1xT2BET

from macapype.utils.utils_tests import load_test_data
from macapype.utils.misc import parse_key

from macapype.utils.utils_nodes import NodeParams, MapNodeParams, ParseParams

from nipype.interfaces.base import traits
import nipype.interfaces.fsl as fsl

data_path = load_test_data("data_test_macaque")

T1_file = op.join(data_path, "sub-Apache_ses-01_T1w.nii")
T2_file = op.join(data_path, "sub-Apache_ses-01_T2w.nii")


def test_NodeParams_init():
    params = {"t_file": T1_file, "t2_file": T2_file, "aT2": True}
    with pytest.raises(AssertionError):
        bet_crop = NodeParams(interface=T1xT2BET(), params=params,  # noqa
                              name="bet_crop")


def test_NodeParams_load_inputs_from_dict():
    params = {"t_file": T1_file, "t2_file": T2_file, "aT2": True}
    bet_crop = NodeParams(interface=T1xT2BET(), name="bet_crop")
    with pytest.raises(AssertionError):
        bet_crop.load_inputs_from_dict(params)


def test_MapNodeParams():
    params = {"crop": {"args": "88 144 14 180 27 103"}}

    crop_bb = MapNodeParams(fsl.ExtractROI(), name='crop_bb',
                            params=parse_key(params, "crop"),
                            iterfield=["in_file"])

    crop_bb.inputs.in_file = [T1_file, T2_file]

    with pytest.raises(ValueError):
        crop_bb.run()


def test_ParseParams():
    params = {
        "sub-01": {
            "ses-01": {
                "node1": {
                    "arg": 1}}}}

    # should return non empty dict
    key = ("sub-01", "ses-01")
    val = ParseParams(params=params, key=key).run().outputs.parsed_params
    assert len(val) == 1

    # should return non empty dict
    key = "sub-01"
    val = ParseParams(params=params, key=key).run().outputs.parsed_params
    assert len(list(val.keys())) == 1

    # should return empty dict
    key = "sub-02"
    val = ParseParams(params=params, key=key).run().outputs.parsed_params
    assert len(val) == 0

    # should return empty dict
    key = "sub-01"
    params = traits.Undefined
    val = ParseParams(params=params, key=key).run().outputs.parsed_params
    assert len(val) == 0



# from .binary_fill_holes import BinaryFillHoles


def test_BinaryFillHoles():

    """
    TODO
    """
    # val = BinaryFillHoles()
    assert True
    pass


def test_split_indexed_mask():

    from macapype.utils.utils_tests import load_test_data, format_template
    from macapype.nodes.segment import split_indexed_mask

    template_name = "NMT_v2.0_asym"
    nmt_dir = load_test_data(name=template_name)
    params_template = format_template(nmt_dir, template_name)

    list_split_files = split_indexed_mask(params_template["template_seg"])

    assert len(list_split_files) != 0


def test_copy_header():
    assert True
    pass


def test_fill_list_vol():

    from macapype.utils.utils_tests import load_test_data, format_template
    from macapype.nodes.segment import fill_list_vol

    import nibabel as nib

    template_name = "MBM_v3.0.1"
    nmt_dir = load_test_data(name=template_name)
    params_template = format_template(nmt_dir, template_name)

    nb_classes = 5
    vol_3c = [params_template["template_gm"],
              params_template["template_wm"],
              params_template["template_csf"]]

    vol_5c = fill_list_vol(vol_3c, nb_classes)

    first_vol_shape = nib.load(vol_5c[0]).get_fdata().shape
    last_vol_shape = nib.load(vol_5c[-1]).get_fdata().shape

    assert len(vol_5c) == nb_classes

    assert first_vol_shape == last_vol_shape


if __name__ == '__main__':
    test_fill_list_vol()



"""
    Gather all full pipelines

"""
import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

from nipype.interfaces import fsl
from nipype.interfaces import ants

from nipype.interfaces.ants.utils import ImageMath

from nipype.interfaces.niftyreg.regutils import RegResample

from ..utils.utils_nodes import NodeParams

from macapype.nodes.correct_bias import T1xT2BiasFieldCorrection
from macapype.nodes.register import IterREGBET

from macapype.nodes.pad import pad_back

from .prepare import (create_short_preparation_pipe,
                      create_short_preparation_T1_pipe)

from .segment import (create_old_segment_pipe,
                      create_native_old_segment_pipe,
                      create_segment_atropos_pipe,
                      create_segment_atropos_seg_pipe,
                      create_mask_from_seg_pipe,
                      create_5tt_pipe)

from .correct_bias import (create_masked_correct_bias_pipe,
                           create_correct_bias_pipe)

from .register import (create_register_NMT_pipe, create_reg_seg_pipe)

from .extract_brain import create_extract_pipe

from .surface import (create_nii_to_mesh_pipe,
                      create_nii2mesh_brain_pipe, create_IsoSurface_brain_pipe)

from macapype.utils.misc import parse_key, list_input_files, show_files

###############################################################################
# SPM based segmentation (from: Régis Trapeau)
# -soft SPM or SPM_T1
###############################################################################


def create_full_spm_subpipes(
        params_template, params_template_stereo,
        params={}, name='full_spm_subpipes',
        mask_file=None, pad=False, space='template'):
    """ Description: SPM based segmentation pipeline from T1w and T2w images
    in template space

    Processing steps:

    - Data preparation (short, with betcrop or crop)
    - debias using T1xT2BiasFieldCorrection (using mask is betcrop)
    - registration to template file with IterREGBET
    - SPM segmentation the old way (SPM8, not dartel based)

    Params:

    - short_preparation_pipe (see :class:`create_short_preparation_pipe \
    <macapype.pipelines.prepare.create_short_preparation_pipe>`)
    - debias (see :class:`T1xT2BiasFieldCorrection \
    <macapype.nodes.correct_bias.T1xT2BiasFieldCorrection>`) - also available \
    as :ref:`indiv_params <indiv_params>`
    - reg (see :class:`IterREGBET <macapype.nodes.register.IterREGBET>`) - \
    also available as :ref:`indiv_params <indiv_params>`
    - native_old_segment_pipe (see :class:`create_native_old_segment_pipe \
    <macapype.pipelines.segment.create_old_segment_pipe>`)

    Inputs:

        inputnode:

            list_T1:
                T1 file names

            list_T2:
                T2 file names

            indiv_params (opt):
                dict with individuals parameters for some nodes


        arguments:

            params_template:
                dict of template files containing brain_template and priors \
            (list of template based segmented tissues)

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "full_spm_subpipes")

    Outputs:

            old_segment_pipe.thresh_gm.out_file:
                segmented grey matter in template space

            old_segment_pipe.thresh_wm.out_file:
                segmented white matter in template space

            old_segment_pipe.thresh_csf.out_file:
                segmented csf in template space
    """

    print("Full pipeline name: ", name)

    # Creating pipeline
    seg_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'native_T1', 'native_T2', 'stereo_T1', 'stereo_T2',
            "stereo_padded_T1",
            "stereo_padded_T2",

            'stereo_brain_mask',
            'stereo_debiased_T1',
            'stereo_debiased_T2',
            'stereo_masked_debiased_T1',
            'stereo_masked_debiased_T2',

            'native_brain_mask', 'native_debiased_T1',
            'native_masked_debiased_T1', 'native_debiased_T2',
            'native_masked_debiased_T2',

            'stereo_prob_wm', 'stereo_prob_gm',
            'stereo_prob_csf',
            'stereo_segmented_brain_mask',
            'stereo_gen_5tt',

            'native_segmented_brain_mask',
            'native_prob_wm', 'native_prob_gm', 'native_prob_csf',
            'native_gen_5tt',

            "stereo_wmgm_mask", "native_wmgm_mask",
            "wmgm_stl",

            'stereo_to_native_trans',
            "native_to_stereo_trans"]),
        name='outputnode')

    # preprocessing
    if 'short_preparation_pipe' in params.keys():
        data_preparation_pipe = create_short_preparation_pipe(
            params=parse_key(params, "short_preparation_pipe"),
            params_template=params_template_stereo)

    else:
        print("Error, short_preparation_pipe, \
            was not found in params, skipping")

        test_node = pe.Node(niu.Function(input_names=['list_T1', 'list_T2'],
                                         output_names=[''],
                                         function=list_input_files),
                            name="test_node")

        seg_pipe.connect(inputnode, 'list_T1',
                         test_node, 'list_T1')
        seg_pipe.connect(inputnode, 'list_T2',
                         test_node, 'list_T2')

        return seg_pipe

    seg_pipe.connect(inputnode, 'list_T1',
                     data_preparation_pipe, 'inputnode.list_T1')
    seg_pipe.connect(inputnode, 'list_T2',
                     data_preparation_pipe, 'inputnode.list_T2')
    seg_pipe.connect(inputnode, 'indiv_params',
                     data_preparation_pipe, 'inputnode.indiv_params')

    # outputnode
    seg_pipe.connect(data_preparation_pipe, 'outputnode.native_T1',
                     outputnode, 'native_T1')

    seg_pipe.connect(data_preparation_pipe, 'outputnode.native_T2',
                     outputnode, 'native_T2')

    if "denoise" in params["short_preparation_pipe"].keys():
        seg_pipe.connect(data_preparation_pipe,
                         'outputnode.stereo_denoised_T1',
                         outputnode, 'stereo_denoised_T1')

        seg_pipe.connect(data_preparation_pipe,
                         'outputnode.stereo_denoised_T2',
                         outputnode, 'stereo_denoised_T2')

    seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_padded_T1",
                     outputnode, "stereo_padded_T1")

    seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_padded_T2",
                     outputnode, "stereo_padded_T2")

    seg_pipe.connect(
        data_preparation_pipe,  "outputnode.stereo_to_native_trans",
        outputnode, 'stereo_to_native_trans')

    seg_pipe.connect(
        data_preparation_pipe,  "outputnode.native_to_stereo_trans",
        outputnode, 'native_to_stereo_trans')

    # debias
    debias = NodeParams(T1xT2BiasFieldCorrection(),
                        params=parse_key(params, "debias"),
                        name='debias')

    if "denoise" in params["short_preparation_pipe"].keys():
        seg_pipe.connect(
            data_preparation_pipe, 'outputnode.stereo_denoised_T1',
            debias, 't1_file')
        seg_pipe.connect(
            data_preparation_pipe, 'outputnode.stereo_denoised_T2',
            debias, 't2_file')
    else:
        seg_pipe.connect(
            data_preparation_pipe, 'outputnode.stereo_T1',
            debias, 't1_file')
        seg_pipe.connect(
            data_preparation_pipe, 'outputnode.stereo_T2',
            debias, 't2_file')

    seg_pipe.connect(inputnode, ('indiv_params', parse_key, "debias"),
                     debias, 'indiv_params')

    if mask_file is None:
        debias.inputs.bet = 1

    else:
        print("Using native external mask {}".format(mask_file))

        # apply transfo to list
        apply_crop_external_mask = pe.Node(RegResample(inter_val="NN"),
                                           name='apply_crop_external_mask')

        apply_crop_external_mask.inputs.flo_file = mask_file

        seg_pipe.connect(data_preparation_pipe,
                         'outputnode.native_to_stereo_trans',
                         apply_crop_external_mask, "trans_file")

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T1",
                apply_crop_external_mask, "ref_file")
        else:
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_T1",
                apply_crop_external_mask, "ref_file")

        seg_pipe.connect(apply_crop_external_mask, "out_file",
                         debias, "b")

    # outputnode

    if mask_file is None:
        seg_pipe.connect(debias, "debiased_mask_file",
                         outputnode, "stereo_brain_mask")
    else:
        seg_pipe.connect(apply_crop_external_mask, "out_file",
                         outputnode, "stereo_brain_mask")

    seg_pipe.connect(debias, 't1_debiased_brain_file',
                     outputnode, "stereo_masked_debiased_T1")

    seg_pipe.connect(debias, 't1_debiased_file',
                     outputnode, "stereo_debiased_T1")

    seg_pipe.connect(debias, 't2_debiased_brain_file',
                     outputnode, "stereo_masked_debiased_T2")

    seg_pipe.connect(debias, 't2_debiased_file',
                     outputnode, "stereo_debiased_T2")

    if pad:
        if mask_file is None:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                debias, "debiased_mask_file",
                outputnode, "native_brain_mask",  params)
        else:
            outputnode.inputs.native_brain_mask = mask_file

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            debias, "t1_debiased_brain_file",
            outputnode, "native_masked_debiased_T1", params, inter_val="LIN")
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            debias, "t1_debiased_file",
            outputnode, "native_debiased_T1", params, inter_val="LIN")
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            debias, "t2_debiased_brain_file",
            outputnode, "native_masked_debiased_T2", params, inter_val="LIN")
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            debias, "t2_debiased_file",
            outputnode, "native_debiased_T2", params, inter_val="LIN")

    # Bias correction of cropped images
    if "reg" not in params.keys():

        print("No reg, skipping")
        return seg_pipe

    # Iterative registration to the INIA19 template
    reg = NodeParams(IterREGBET(),
                     params=parse_key(params, "reg"),
                     name='reg')

    reg.inputs.refb_file = params_template["template_brain"]

    seg_pipe.connect(debias, 't1_debiased_file',
                     reg, 'inw_file')

    seg_pipe.connect(debias, 't1_debiased_brain_file',
                     reg, 'inb_file')

    seg_pipe.connect(inputnode, ('indiv_params', parse_key, "reg"),
                     reg, 'indiv_params')

    # Compute brain mask using old_segment of SPM and postprocessing on
    # tissues' masks
    if "old_segment_pipe" not in params.keys():
        print("No segmentation, skipping")
        return seg_pipe

    if space == "template":

        old_segment_pipe = create_old_segment_pipe(
            params_template, params=parse_key(params, "old_segment_pipe"))

        seg_pipe.connect(reg, 'warp_file',
                         old_segment_pipe, 'inputnode.T1')

        seg_pipe.connect(inputnode, 'indiv_params',
                         old_segment_pipe, 'inputnode.indiv_params')

    elif space == "native":

        old_segment_pipe = create_native_old_segment_pipe(
            params_template, params=parse_key(params, "old_segment_pipe"))

        seg_pipe.connect(reg, 'inv_transfo_file',
                         old_segment_pipe, 'inputnode.inv_transfo_file')

        seg_pipe.connect(debias, 't1_debiased_brain_file',
                         old_segment_pipe, 'inputnode.native_T1')

        seg_pipe.connect(inputnode, 'indiv_params',
                         old_segment_pipe, 'inputnode.indiv_params')
    else:

        print("Error, space={}".format(space))
        return seg_pipe

    # outputnode
    seg_pipe.connect(
        old_segment_pipe, 'outputnode.prob_wm',
        outputnode, 'stereo_prob_wm')

    seg_pipe.connect(
        old_segment_pipe, 'outputnode.prob_csf',
        outputnode, 'stereo_prob_csf')

    seg_pipe.connect(
        old_segment_pipe, 'outputnode.prob_gm',
        outputnode, 'stereo_prob_gm')

    if pad and space == "native":

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            old_segment_pipe, "outputnode.prob_gm",
            outputnode, "native_prob_gm", params, inter_val="LIN")

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            old_segment_pipe, "outputnode.prob_wm",
            outputnode, "native_prob_wm", params, inter_val="LIN")

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            old_segment_pipe, "outputnode.prob_csf",
            outputnode, "native_prob_csf", params, inter_val="LIN")

    if "mask_from_seg_pipe" in params.keys():

        mask_from_seg_pipe = create_mask_from_seg_pipe(
            params=parse_key(params, "mask_from_seg_pipe"))

        seg_pipe.connect(old_segment_pipe, 'outputnode.threshold_csf',
                         mask_from_seg_pipe, 'inputnode.mask_csf')

        seg_pipe.connect(old_segment_pipe, 'outputnode.threshold_wm',
                         mask_from_seg_pipe, 'inputnode.mask_wm')

        seg_pipe.connect(old_segment_pipe, 'outputnode.threshold_gm',
                         mask_from_seg_pipe, 'inputnode.mask_gm')

        seg_pipe.connect(inputnode, 'indiv_params',
                         mask_from_seg_pipe, 'inputnode.indiv_params')

        # outputnode
        seg_pipe.connect(mask_from_seg_pipe, "wmgm2mesh.stl_file",
                         outputnode, 'wmgm_stl')

        seg_pipe.connect(
            mask_from_seg_pipe,
            'merge_indexed_mask.indexed_mask',
            outputnode, 'stereo_segmented_brain_mask')

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                mask_from_seg_pipe,
                'merge_indexed_mask.indexed_mask',
                outputnode, "native_segmented_brain_mask", params)

    if "export_5tt_pipe" in params["old_segment_pipe"].keys():

        export_5tt_pipe = create_5tt_pipe(
            params=parse_key(params["old_segment_pipe"], "export_5tt_pipe"))

        seg_pipe.connect(
            old_segment_pipe, 'outputnode.threshold_gm',
            export_5tt_pipe, 'inputnode.gm_file')

        seg_pipe.connect(
            old_segment_pipe, 'outputnode.threshold_wm',
            export_5tt_pipe, 'inputnode.wm_file')

        seg_pipe.connect(
            old_segment_pipe, 'outputnode.threshold_csf',
            export_5tt_pipe, 'inputnode.csf_file')

        seg_pipe.connect(
            export_5tt_pipe, 'export_5tt.gen_5tt_file',
            outputnode, 'stereo_gen_5tt')

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                export_5tt_pipe, 'export_5tt.gen_5tt_file',
                outputnode, "native_gen_5tt", params)

    return seg_pipe

###############################################################################
# ANTS based segmentation (from Kepkee Loh / Julien Sein)
# (-soft ANTS)
###############################################################################


def create_brain_segment_from_mask_pipe(
        params_template, params={}, name="brain_segment_from_mask_pipe",
        space="native"):
    """ Description: Segment T1 (using T2 for bias correction) and a previously
    computed mask with NMT Atlas and atropos segment.

    Params:

    - masked_correct_bias_pipe (see :class:`create_masked_correct_bias_pipe \
    <macapype.pipelines.correct_bias.create_masked_correct_bias_pipe>`)
    - register_NMT_pipe (see :class:`create_register_NMT_pipe \
    <macapype.pipelines.register.create_register_NMT_pipe>`)
    - segment_atropos_pipe (see :class:`create_segment_atropos_pipe \
    <macapype.pipelines.segment.create_segment_atropos_pipe>`)

    Inputs:

        inputnode:

            preproc_T1:
                preprocessed T1 file name

            preproc_T2:
                preprocessed T2 file name

            brain_mask:
                a mask computed for the same T1/T2 images

            indiv_params (opt):
                dict with individuals parameters for some nodes


        arguments:

            params_template:
                dictionary of template files

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "full_segment_pipe")

    Outputs:

    """
    # creating pipeline
    brain_segment_pipe = pe.Workflow(name=name)

    # creating inputnode
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['masked_debiased_T1', "debiased_T1", 'indiv_params']),
        name='inputnode')

    # creating outputnode
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=["segmented_file", "threshold_gm", "threshold_wm",
                    "threshold_csf", "prob_gm", "prob_wm",
                    "prob_csf", "gen_5tt"]),
        name='outputnode')

    if "use_priors" in params["segment_atropos_pipe"].keys():

        if "register_NMT_pipe" in params:

            # register NMT template, template mask and priors to subject T1
            register_NMT_pipe = create_register_NMT_pipe(
                params_template=params_template,
                params=parse_key(params, "register_NMT_pipe"))

            brain_segment_pipe.connect(
                inputnode, 'masked_debiased_T1',
                register_NMT_pipe, "inputnode.T1")

            brain_segment_pipe.connect(
                inputnode, 'indiv_params',
                register_NMT_pipe, "inputnode.indiv_params")

        elif "reg" in params:
            # Iterative registration to the template
            reg = NodeParams(
                IterREGBET(),
                params=parse_key(params, "reg"),
                name='reg')

            reg.inputs.refb_file = params_template["template_brain"]

            if "nonlin_reg" in params["reg"]:
                reg.inputs.refw_file = params_template["template_head"]
                reg.inputs.k = True

            brain_segment_pipe.connect(
                inputnode, 'debiased_T1',
                reg, 'inw_file')

            brain_segment_pipe.connect(
                inputnode, 'masked_debiased_T1',
                reg, 'inb_file')

            brain_segment_pipe.connect(
                inputnode, ('indiv_params', parse_key, "reg"),
                reg, 'indiv_params')

            if "template_seg" in params_template.keys():

                # seg
                register_seg_to_nat = pe.Node(
                    fsl.ApplyXFM(), name="register_seg_to_nat")
                register_seg_to_nat.inputs.interp = "nearestneighbour"

                register_seg_to_nat.inputs.in_file = params_template[
                    "template_seg"]
                brain_segment_pipe.connect(
                    inputnode, 'masked_debiased_T1',
                    register_seg_to_nat, 'reference')

                brain_segment_pipe.connect(
                    reg, 'inv_transfo_file',
                    register_seg_to_nat, "in_matrix_file")

            else:
                # gm
                register_gm_to_nat = pe.Node(
                    fsl.ApplyXFM(), name="register_gm_to_nat")
                register_gm_to_nat.inputs.output_type = "NIFTI_GZ"  # for SPM
                register_gm_to_nat.inputs.interp = "nearestneighbour"

                register_gm_to_nat.inputs.in_file = \
                    params_template["template_gm"]

                brain_segment_pipe.connect(
                    inputnode, 'masked_debiased_T1',
                    register_gm_to_nat, 'reference')

                brain_segment_pipe.connect(
                    reg, 'inv_transfo_file',
                    register_gm_to_nat, "in_matrix_file")

                # wm
                register_wm_to_nat = pe.Node(
                    fsl.ApplyXFM(), name="register_wm_to_nat")
                register_wm_to_nat.inputs.output_type = "NIFTI_GZ"  # for SPM
                register_wm_to_nat.inputs.interp = "nearestneighbour"

                register_wm_to_nat.inputs.in_file = \
                    params_template["template_wm"]

                brain_segment_pipe.connect(
                    inputnode, 'masked_debiased_T1',
                    register_wm_to_nat, 'reference')

                brain_segment_pipe.connect(
                    reg, 'inv_transfo_file',
                    register_wm_to_nat, "in_matrix_file")

                # csf
                register_csf_to_nat = pe.Node(
                    fsl.ApplyXFM(), name="register_csf_to_nat")
                register_csf_to_nat.inputs.output_type = "NIFTI_GZ"  # for SPM
                register_csf_to_nat.inputs.interp = "nearestneighbour"

                register_csf_to_nat.inputs.in_file = \
                    params_template["template_csf"]

                brain_segment_pipe.connect(
                    inputnode, 'masked_debiased_T1',
                    register_csf_to_nat, 'reference')

                brain_segment_pipe.connect(
                    reg, 'inv_transfo_file',
                    register_csf_to_nat, "in_matrix_file")
        else:
            print("##### Error, no coregistration method is defined")
            return brain_segment_pipe

    # ants Atropos
    if "template_seg" in params_template.keys():

        print("#### create_segment_atropos_seg_pipe ")
        segment_atropos_pipe = create_segment_atropos_seg_pipe(
            params=parse_key(params, "segment_atropos_pipe"))

        # linking priors if "use_priors" in params
        if "use_priors" in params["segment_atropos_pipe"].keys():

            if "register_NMT_pipe" in params:
                brain_segment_pipe.connect(
                    register_NMT_pipe, 'outputnode.native_template_seg',
                    segment_atropos_pipe, "inputnode.seg_file")
            elif "reg" in params:
                brain_segment_pipe.connect(
                    register_seg_to_nat, 'out_file',
                    segment_atropos_pipe, "inputnode.seg_file")
    else:
        print("#### create_segment_atropos_pipe (3 tissues) ")

        segment_atropos_pipe = create_segment_atropos_pipe(
            params=parse_key(params, "segment_atropos_pipe"))

        # linking priors if "use_priors" in params
        if "use_priors" in params["segment_atropos_pipe"].keys():

            if "register_NMT_pipe" in params:
                brain_segment_pipe.connect(
                    register_NMT_pipe, 'outputnode.native_template_csf',
                    segment_atropos_pipe, "inputnode.csf_prior_file")
                brain_segment_pipe.connect(
                    register_NMT_pipe,  'outputnode.native_template_gm',
                    segment_atropos_pipe, "inputnode.gm_prior_file")
                brain_segment_pipe.connect(
                    register_NMT_pipe,  'outputnode.native_template_wm',
                    segment_atropos_pipe, "inputnode.wm_prior_file")

            elif "reg" in params:
                print("To be validated...")
                brain_segment_pipe.connect(
                    register_csf_to_nat, "out_file",
                    segment_atropos_pipe, "inputnode.csf_prior_file")
                brain_segment_pipe.connect(
                    register_gm_to_nat, "out_file",
                    segment_atropos_pipe, "inputnode.gm_prior_file")
                brain_segment_pipe.connect(
                    register_wm_to_nat, "out_file",
                    segment_atropos_pipe, "inputnode.wm_prior_file")

    # input to segment_atropos_pipe
    brain_segment_pipe.connect(
        inputnode, 'masked_debiased_T1',
        segment_atropos_pipe, "inputnode.brain_file")

    if "export_5tt_pipe" in params.keys():

        export_5tt_pipe = create_5tt_pipe(
            params=parse_key(params, "export_5tt_pipe"))

        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_gm',
                                   export_5tt_pipe, 'inputnode.gm_file')

        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_wm',
                                   export_5tt_pipe, 'inputnode.wm_file')

        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_csf',
                                   export_5tt_pipe, 'inputnode.csf_file')

        brain_segment_pipe.connect(export_5tt_pipe, 'export_5tt.gen_5tt_file',
                                   outputnode, 'gen_5tt')

    if space == 'native':

        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.segmented_file',
                                   outputnode, 'segmented_file')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_gm',
                                   outputnode, 'threshold_gm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_wm',
                                   outputnode, 'threshold_wm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_csf',
                                   outputnode, 'threshold_csf')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_gm',
                                   outputnode, 'prob_gm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_wm',
                                   outputnode, 'prob_wm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_csf',
                                   outputnode, 'prob_csf')

    elif space == "template":
        reg_seg_pipe = create_reg_seg_pipe()

        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.segmented_file',
                                   reg_seg_pipe,
                                   'inputnode.native_seg')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_gm',
                                   reg_seg_pipe,
                                   'inputnode.native_threshold_gm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_wm',
                                   reg_seg_pipe,
                                   'inputnode.native_threshold_wm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.threshold_csf',
                                   reg_seg_pipe,
                                   'inputnode.native_threshold_csf')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_gm',
                                   reg_seg_pipe,
                                   'inputnode.native_prob_gm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_wm',
                                   reg_seg_pipe,
                                   'inputnode.native_prob_wm')
        brain_segment_pipe.connect(segment_atropos_pipe,
                                   'outputnode.prob_csf',
                                   reg_seg_pipe,
                                   'inputnode.native_prob_csf')

        # other inputs
        if "reg" in params:
            brain_segment_pipe.connect(
                reg, 'transfo_file',
                reg_seg_pipe, 'inputnode.transfo_file')

        elif "register_NMT_pipe" in params:
            brain_segment_pipe.connect(
                register_NMT_pipe,
                'NMT_subject_align.transfo_file',
                reg_seg_pipe, 'inputnode.transfo_file')

        reg_seg_pipe.inputs.inputnode.ref_image = \
            params_template['template_head']

        # output node
        brain_segment_pipe.connect(reg_seg_pipe, 'outputnode.norm_seg',
                                   outputnode, 'segmented_file')
        brain_segment_pipe.connect(reg_seg_pipe,
                                   'outputnode.norm_threshold_gm',
                                   outputnode, 'threshold_gm')
        brain_segment_pipe.connect(reg_seg_pipe,
                                   'outputnode.norm_threshold_wm',
                                   outputnode, 'threshold_wm')
        brain_segment_pipe.connect(reg_seg_pipe,
                                   'outputnode.norm_threshold_csf',
                                   outputnode, 'threshold_csf')
        brain_segment_pipe.connect(reg_seg_pipe, 'outputnode.norm_prob_gm',
                                   outputnode, 'prob_gm')
        brain_segment_pipe.connect(reg_seg_pipe, 'outputnode.norm_prob_wm',
                                   outputnode, 'prob_wm')
        brain_segment_pipe.connect(reg_seg_pipe, 'outputnode.norm_prob_csf',
                                   outputnode, 'prob_csf')

    return brain_segment_pipe

# #############################################################################################################################


def create_full_ants_subpipes(
        params_template, params_template_stereo,
        params={}, name="full_ants_subpipes", mask_file=None,
        space="native", pad=False):
    """Description: Segment T1 (using T2 for bias correction) .

    Params:

    - short_preparation_pipe (see :class:`create_short_preparation_pipe \
    <macapype.pipelines.prepare.create_short_preparation_pipe>`)

    - brain_extraction_pipe (see :class:`create_brain_extraction_pipe \
    <macapype.pipelines.full_pipelines.create_brain_extraction_pipe>`)
    - brain_segment_pipe (see :class:`create_brain_segment_from_mask_pipe\
    <macapype.pipelines.full_pipelines.create_brain_segment_from_mask_pipe>`)
    - nii_to_mesh_pipe (see :class:`create_nii_to_mesh_pipe\
    <macapype.pipelines.surface.create_nii_to_mesh_pipe>`)

    Inputs:

        inputnode:

            list_T1:
                preprocessed T1 file name

            list_T2:
                preprocessed T2 file name

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params_template:
                dictionary of template files

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "full_segment_pipe")

    Outputs:

        brain_mask

        segmented_brain_mask:
            indexed with tissue classes

    """

    # creating pipeline
    seg_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # output node
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                    "native_T1", "native_T2",
                    'stereo_T1', 'stereo_T2',
                    "stereo_padded_T1",
                    "stereo_padded_T2",

                    "stereo_denoised_T1",
                    "stereo_denoised_T2",

                    'stereo_debiased_T1', 'stereo_debiased_T2',
                    "native_debiased_T1", "native_debiased_T2",

                    'stereo_brain_mask',
                    'stereo_padded_brain_mask',
                    'stereo_masked_debiased_T1',
                    'stereo_masked_debiased_T2',

                    'native_brain_mask', "native_masked_debiased_T1",
                    "native_masked_debiased_T2",

                    'stereo_segmented_brain_mask',
                    'stereo_padded_segmented_brain_mask',
                    'stereo_prob_gm', 'stereo_prob_wm', 'stereo_prob_csf',
                    "stereo_gen_5tt",

                    'native_segmented_brain_mask',
                    'native_prob_gm', 'native_prob_wm', 'native_prob_csf',
                    "native_gen_5tt",

                    "stereo_wmgm_mask",
                    "native_wmgm_mask",
                    "wmgm_stl",

                    "stereo_to_native_trans",
                    "native_to_stereo_trans"]),
        name='outputnode')

    # preprocessing
    if 'short_preparation_pipe' in params.keys():
        data_preparation_pipe = create_short_preparation_pipe(
            params=parse_key(params, "short_preparation_pipe"),
            params_template=params_template_stereo)

    else:
        print("Error, short_preparation_pipe \
            was not found in params, skipping")

        test_node = pe.Node(niu.Function(input_names=['list_T1', 'list_T2'],
                                         output_names=[''],
                                         function=list_input_files),
                            name="test_node")

        seg_pipe.connect(inputnode, 'list_T1',
                         test_node, 'list_T1')
        seg_pipe.connect(inputnode, 'list_T2',
                         test_node, 'list_T2')

        return seg_pipe

    # inputs
    seg_pipe.connect(inputnode, 'list_T1',
                     data_preparation_pipe, 'inputnode.list_T1')

    seg_pipe.connect(inputnode, 'list_T2',
                     data_preparation_pipe, 'inputnode.list_T2')

    seg_pipe.connect(inputnode, 'indiv_params',
                     data_preparation_pipe, 'inputnode.indiv_params')

    # outputnode
    seg_pipe.connect(data_preparation_pipe, 'outputnode.native_T1',
                     outputnode, 'native_T1')

    seg_pipe.connect(data_preparation_pipe, 'outputnode.native_T2',
                     outputnode, 'native_T2')

    # everything is now in stereo space

    if "denoise" in params["short_preparation_pipe"].keys():
        seg_pipe.connect(
            data_preparation_pipe, "outputnode.stereo_denoised_T1",
            outputnode, "stereo_denoised_T1")
        seg_pipe.connect(
            data_preparation_pipe, "outputnode.stereo_denoised_T2",
            outputnode, "stereo_denoised_T2")

    seg_pipe.connect(data_preparation_pipe, 'outputnode.stereo_T1',
                     outputnode, "stereo_T1")

    seg_pipe.connect(data_preparation_pipe, 'outputnode.stereo_T2',
                     outputnode, "stereo_T2")

    seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_padded_T1",
                     outputnode, "stereo_padded_T1")

    seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_padded_T2",
                     outputnode, "stereo_padded_T2")

    seg_pipe.connect(
        data_preparation_pipe, "outputnode.stereo_to_native_trans",
        outputnode, 'stereo_to_native_trans')

    seg_pipe.connect(
        data_preparation_pipe, "outputnode.native_to_stereo_trans",
        outputnode, 'native_to_stereo_trans')

    # ######################################## correct bias
    assert not ("fast" in params.keys() and "N4debias" in
                params.keys()), "error, only one of correct_bias_pipe\
                or N4debias should be present"

    if "correct_bias_pipe" in params.keys():

        print("Found create_correct_bias_pipe in params.json")

        # Correct_bias_T1_T2
        correct_bias_pipe = create_correct_bias_pipe(
            params=parse_key(params, "correct_bias_pipe"))

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T1",
                correct_bias_pipe, 'inputnode.preproc_T1')
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T2",
                correct_bias_pipe, 'inputnode.preproc_T1')
        else:
            seg_pipe.connect(data_preparation_pipe, 'outputnode.stereo_T1',
                             correct_bias_pipe, 'inputnode.preproc_T1')

            seg_pipe.connect(data_preparation_pipe, 'outputnode.stereo_T2',
                             correct_bias_pipe, 'inputnode.preproc_T2')

        # outputnode
        seg_pipe.connect(correct_bias_pipe,
                         "outputnode.debiased_T1",
                         outputnode, "stereo_debiased_T1")

        seg_pipe.connect(correct_bias_pipe,
                         "outputnode.debiased_T2",
                         outputnode, "stereo_debiased_T2")

        if pad:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                correct_bias_pipe, "outputnode.debiased_T1",
                outputnode, "native_debiased_T1", params, inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                correct_bias_pipe, "outputnode.debiased_T2",
                outputnode, "native_debiased_T2", params, inter_val="LIN")

    elif "N4debias" in params.keys():
        print("Found N4debias in params.json")

        # N4 intensity normalization over T1
        N4debias_T1 = NodeParams(ants.N4BiasFieldCorrection(),
                                 params=parse_key(params, "N4debias"),
                                 name='N4debias_T1')

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T1",
                N4debias_T1, "input_image")
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                N4debias_T1, "input_image")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "N4debias"),
            N4debias_T1, "indiv_params")

        # N4 intensity normalization over T2
        N4debias_T2 = NodeParams(ants.N4BiasFieldCorrection(),
                                 params=parse_key(params, "N4debias"),
                                 name='N4debias_T2')

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T2",
                N4debias_T2, "input_image")
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T2',
                N4debias_T2, "input_image")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "N4debias"),
            N4debias_T2, "indiv_params")

        # outputnode
        seg_pipe.connect(
            N4debias_T1, "output_image",
            outputnode, "stereo_debiased_T1")

        seg_pipe.connect(
            N4debias_T2, "output_image",
            outputnode, "stereo_debiased_T2")

        if pad:

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                N4debias_T1, "output_image",
                outputnode, "native_debiased_T1", params, inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                N4debias_T2, "output_image",
                outputnode, "native_debiased_T2", params, inter_val="LIN")

    elif "fast" in params.keys():

        print("Found fast in params.json")

        # fast over T1
        fast_T1 = NodeParams(
            fsl.FAST(),
            params=parse_key(params, "fast"),
            name='fast_T1')

        fast_T1.inputs.output_biascorrected = True
        fast_T1.inputs.output_biasfield = True
        fast_T1.inputs.img_type = 1

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T1",
                fast_T1, "in_files")
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                fast_T1, "in_files")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "fast"),
            fast_T1, "indiv_params")

        # fast over T2
        fast_T2 = NodeParams(
            fsl.FAST(),
            params=parse_key(params, "fast"),
            name='fast_T2')

        fast_T2.inputs.output_biascorrected = True
        fast_T2.inputs.output_biasfield = True
        fast_T2.inputs.img_type = 2

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, "outputnode.stereo_denoised_T2",
                fast_T2, "in_files")
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T2',
                fast_T2, "in_files")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "fast"),
            fast_T2, "indiv_params")

        # outputnode
        seg_pipe.connect(
            fast_T1, "restored_image",
            outputnode, "stereo_debiased_T1")

        seg_pipe.connect(
            fast_T2, "restored_image",
            outputnode, "stereo_debiased_T2")

        if pad:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                fast_T1, "restored_image",
                outputnode, "native_debiased_T1", params, inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                fast_T2, "restored_image",
                outputnode, "native_debiased_T2", params, inter_val="LIN")

    else:
        print("No debias will be performed before extract_pipe")

    # ################# extract mask
    print("mask file {}".format(mask_file))

    if mask_file is None:

        # full extract brain pipeline (correct_bias, denoising, extract brain)
        if "extract_pipe" in params.keys():

            # brain extraction
            extract_pipe = create_extract_pipe(
                params_template=params_template,
                params=parse_key(params, "extract_pipe"))

            seg_pipe.connect(inputnode, "indiv_params",
                                        extract_pipe, "inputnode.indiv_params")

            if "use_T2" in params["extract_pipe"]:

                if "correct_bias_pipe" in params:
                    seg_pipe.connect(
                        correct_bias_pipe, "outputnode.debiased_T2",
                        extract_pipe, "inputnode.restore_T1")

                elif "N4debias" in params.keys():
                    # brain extraction
                    seg_pipe.connect(
                        N4debias_T2, "output_image",
                        extract_pipe, "inputnode.restore_T1")

                elif "fast" in params.keys():

                    # brain extraction
                    seg_pipe.connect(
                        fast_T2, ("restored_image", show_files),
                        extract_pipe, "inputnode.restore_T1")
                else:

                    if "denoise" in params["short_preparation_pipe"].keys():
                        seg_pipe.connect(
                            data_preparation_pipe,
                            "outputnode.stereo_denoised_T2",
                            extract_pipe, "inputnode.restore_T1")
                    else:
                        seg_pipe.connect(
                            data_preparation_pipe,
                            'outputnode.stereo_T2',
                            extract_pipe, "inputnode.restore_T1")

            else:

                if "correct_bias_pipe" in params:
                    seg_pipe.connect(
                        correct_bias_pipe, "outputnode.debiased_T1",
                        extract_pipe, "inputnode.restore_T1")

                elif "N4debias" in params.keys():
                    # brain extraction
                    seg_pipe.connect(
                        N4debias_T1, "output_image",
                        extract_pipe, "inputnode.restore_T1")

                elif "fast" in params.keys():

                    # brain extraction
                    seg_pipe.connect(
                        fast_T1, ("restored_image", show_files),
                        extract_pipe, "inputnode.restore_T1")
                else:

                    if "denoise" in params["short_preparation_pipe"].keys():
                        seg_pipe.connect(
                            data_preparation_pipe,
                            "outputnode.stereo_denoised_T1",
                            extract_pipe, "inputnode.restore_T1")
                    else:
                        seg_pipe.connect(
                            data_preparation_pipe,
                            'outputnode.stereo_T1',
                            extract_pipe, "inputnode.restore_T1")

            # outputnode
            seg_pipe.connect(
                extract_pipe, "smooth_mask.out_file",
                outputnode, "stereo_brain_mask")

            if "pad_template" in params["short_preparation_pipe"].keys():

                pad_stereo_brain_mask = NodeParams(
                    ImageMath(),
                    params=parse_key(params["short_preparation_pipe"],
                                     "pad_template"),
                    name="pad_stereo_brain_mask")

                seg_pipe.connect(
                    extract_pipe, "smooth_mask.out_file",
                    pad_stereo_brain_mask, "op1")

                seg_pipe.connect(
                    pad_stereo_brain_mask, "output_image",
                    outputnode, "stereo_padded_brain_mask")

            if pad:
                pad_back(
                    seg_pipe, data_preparation_pipe, inputnode,
                    extract_pipe, "smooth_mask.out_file",
                    outputnode, "native_brain_mask", params)

        # full extract brain pipeline (correct_bias, denoising, extract brain)
        else:
            print("no extract_brain method is defined, skipping")
            return seg_pipe

    else:
        print("Using native external mask {}".format(mask_file))
        outputnode.inputs.native_brain_mask = mask_file

        # apply transfo to list
        apply_crop_external_mask = pe.Node(RegResample(inter_val="NN"),
                                           name='apply_crop_external_mask')

        apply_crop_external_mask.inputs.flo_file = mask_file

        seg_pipe.connect(data_preparation_pipe,
                         'outputnode.native_to_stereo_trans',
                         apply_crop_external_mask, "trans_file")

        seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_T1",
                         apply_crop_external_mask, "ref_file")

        # outputnode
        seg_pipe.connect(apply_crop_external_mask, "out_file",
                         outputnode, "stereo_brain_mask")

    # ################################################ masked_debias ##
    # correcting for bias T1/T2, but this time with a mask
    if "masked_correct_bias_pipe" in params.keys():

        masked_correct_bias_pipe = create_masked_correct_bias_pipe(
            params=parse_key(params, "masked_correct_bias_pipe"))

        if "correct_bias_pipe" in params.keys():

            seg_pipe.connect(correct_bias_pipe, "outputnode.debiased_T1",
                             masked_correct_bias_pipe, "inputnode.preproc_T1")

            seg_pipe.connect(correct_bias_pipe, "outputnode.debiased_T2",
                             masked_correct_bias_pipe, "inputnode.preproc_T2")

        elif "N4debias" in params.keys():

            seg_pipe.connect(N4debias_T1, "output_image",
                             masked_correct_bias_pipe, "inputnode.preproc_T1")

            seg_pipe.connect(N4debias_T2, "output_image",
                             masked_correct_bias_pipe, "inputnode.preproc_T2")

        elif "fast" in params.keys():

            seg_pipe.connect(fast_T1, ("restored_image", show_files),
                             masked_correct_bias_pipe, "inputnode.preproc_T1")

            seg_pipe.connect(fast_T2, ("restored_image", show_files),
                             masked_correct_bias_pipe, "inputnode.preproc_T2")

        else:
            if "denoise" in params["short_preparation_pipe"].keys():
                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                    masked_correct_bias_pipe, "inputnode.preproc_T1")

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T2',
                    masked_correct_bias_pipe, "inputnode.preproc_T2")
            else:

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T1',
                    masked_correct_bias_pipe, "inputnode.preproc_T1")

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T2',
                    masked_correct_bias_pipe, "inputnode.preproc_T2")

        if mask_file is None:
            seg_pipe.connect(extract_pipe, "smooth_mask.out_file",
                             masked_correct_bias_pipe, "inputnode.brain_mask")

        else:
            seg_pipe.connect(apply_crop_external_mask, "out_file",
                             masked_correct_bias_pipe, "inputnode.brain_mask")

        # outputnode
        seg_pipe.connect(
            masked_correct_bias_pipe,
            'outputnode.mask_debiased_T1',
            outputnode, "stereo_masked_debiased_T1")

        seg_pipe.connect(
            masked_correct_bias_pipe,
            'outputnode.mask_debiased_T2',
            outputnode, "stereo_masked_debiased_T1")

        if pad:

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                masked_correct_bias_pipe,
                'outputnode.mask_debiased_T1',
                outputnode, "native_masked_debiased_T1", params,
                inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                masked_correct_bias_pipe,
                'outputnode.mask_debiased_T2',
                outputnode, "native_masked_debiased_T2", params,
                inter_val="LIN")

    elif "debias" in params.keys():
        # Bias correction of cropped images
        debias = NodeParams(T1xT2BiasFieldCorrection(),
                            params=parse_key(params, "debias"),
                            name='debias')

        if "N4debias" in params.keys():

            seg_pipe.connect(N4debias_T1, "output_image",
                             debias, 't1_file')

            seg_pipe.connect(N4debias_T2, "output_image",
                             debias, 't2_file')

        elif "fast" in params.keys():

            seg_pipe.connect(fast_T1, ("restored_image", show_files),
                             debias, 't1_file')

            seg_pipe.connect(fast_T2, ("restored_image", show_files),
                             debias, 't2_file')

        else:
            if "denoise" in params["short_preparation_pipe"].keys():
                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                    debias, 't1_file')

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T2',
                    debias, 't2_file')
            else:

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T1',
                    debias, 't1_file')

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T2',
                    debias, 't2_file')

        if mask_file is None:
            if "extract_pipe" in params.keys():
                seg_pipe.connect(
                    extract_pipe, "smooth_mask.out_file",
                    debias, 'b')
            else:
                debias.inputs.bet = 1

                # outputnode
                seg_pipe.connect(
                    debias, "debiased_mask_file",
                    outputnode, "stereo_brain_mask")

                if pad:
                    pad_back(
                        seg_pipe, data_preparation_pipe, inputnode,
                        debias, "debiased_mask_file",
                        outputnode, "native_brain_mask", params)

        else:
            seg_pipe.connect(
                apply_crop_external_mask, "out_file",
                debias, 'b')

        # TODO is not used now...
        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "debias"),
            debias, 'indiv_params')

        # outputnode
        seg_pipe.connect(
            debias, 't1_debiased_brain_file',
            outputnode, "stereo_masked_debiased_T1")

        seg_pipe.connect(
            debias, 't2_debiased_brain_file',
            outputnode, "stereo_masked_debiased_T2")

        if pad:

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                debias, 't1_debiased_brain_file',
                outputnode, "native_masked_debiased_T1", params,
                inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                debias, 't2_debiased_brain_file',
                outputnode, "native_masked_debiased_T2", params,
                inter_val="LIN")

    else:

        print("**** Error, masked_correct_bias_pipe or debias \
            should be in brain_extraction_pipe of params.json")
        print("No T1*T2 debias will be performed")

        # restore_mask_T1
        restore_mask_T1 = pe.Node(fsl.ApplyMask(), name='restore_mask_T1')

        # restore_mask_T2
        restore_mask_T2 = pe.Node(fsl.ApplyMask(), name='restore_mask_T2')

        if "N4debias" in params.keys():

            seg_pipe.connect(N4debias_T1, "output_image",
                             restore_mask_T1, 'in_file')

            seg_pipe.connect(N4debias_T2, "output_image",
                             restore_mask_T2, 'in_file')

        elif "fast" in params.keys():

            seg_pipe.connect(fast_T1, ("restored_image", show_files),
                             restore_mask_T1, 'in_file')

            seg_pipe.connect(fast_T2, ("restored_image", show_files),
                             restore_mask_T2, 'in_file')

        else:
            if "denoise" in params["short_preparation_pipe"].keys():
                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                    restore_mask_T1, 'in_file')

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T2',
                    restore_mask_T2, 'in_file')
            else:

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T1',
                    restore_mask_T1, 'in_file')

                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T2',
                    restore_mask_T2, 'in_file')

        if mask_file is None:
            if "extract_pipe" in params.keys():
                seg_pipe.connect(
                    extract_pipe, "smooth_mask.out_file",
                    restore_mask_T1, 'mask_file')

                seg_pipe.connect(
                    extract_pipe, "smooth_mask.out_file",
                    restore_mask_T2, 'mask_file')
        else:
            seg_pipe.connect(
                apply_crop_external_mask, "out_file",
                restore_mask_T1, 'mask_file')

            seg_pipe.connect(
                apply_crop_external_mask, "out_file",
                restore_mask_T2, 'mask_file')

        # outputnode
        seg_pipe.connect(
            restore_mask_T1, 'out_file',
            outputnode, "stereo_masked_debiased_T1")

        seg_pipe.connect(
            restore_mask_T2, 'out_file',
            outputnode, "stereo_masked_debiased_T2")

        if pad:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                restore_mask_T1, 'out_file',
                outputnode, "native_masked_debiased_T1", params,
                inter_val="LIN")

            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                restore_mask_T2, 'out_file',
                outputnode, "native_masked_debiased_T2", params,
                inter_val="LIN")

    # ################################### brain_segment
    # (restarting from the avg_align files)
    if "brain_segment_pipe" not in params.keys():
        return seg_pipe

    brain_segment_pipe = create_brain_segment_from_mask_pipe(
        params_template=params_template,
        params=parse_key(params, "brain_segment_pipe"), space=space)

    seg_pipe.connect(inputnode, 'indiv_params',
                     brain_segment_pipe, 'inputnode.indiv_params')

    if 'use_T2' in params['brain_segment_pipe'].keys():

        # using T2
        if "masked_correct_bias_pipe" in params.keys():
            seg_pipe.connect(
                masked_correct_bias_pipe, 'outputnode.mask_debiased_T2',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

            if "correct_bias_pipe" in params.keys():
                seg_pipe.connect(
                    correct_bias_pipe, "outputnode.debiased_T2",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "N4debias" in params.keys():
                seg_pipe.connect(
                    N4debias_T2, "output_image",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "fast" in params.keys():
                seg_pipe.connect(
                    fast_T2, ("restored_image", show_files),
                    brain_segment_pipe, 'inputnode.debiased_T1')

            else:

                if "denoise" in params["short_preparation_pipe"].keys():
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_denoised_T2',
                        brain_segment_pipe, 'inputnode.debiased_T1')

                else:
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_T2',
                        brain_segment_pipe, 'inputnode.debiased_T1')

        elif "debias" in params.keys():
            seg_pipe.connect(
                debias, 't2_debiased_brain_file',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

            seg_pipe.connect(
                debias, 't2_debiased_file',
                brain_segment_pipe, 'inputnode.debiased_T1')

        else:

            if "correct_bias_pipe" in params.keys():
                seg_pipe.connect(
                    correct_bias_pipe, "outputnode.debiased_T2",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "N4debias" in params.keys():
                seg_pipe.connect(
                    N4debias_T2, "output_image",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "fast" in params.keys():
                seg_pipe.connect(
                    fast_T2, ("restored_image", show_files),
                    brain_segment_pipe, 'inputnode.debiased_T1')

            else:

                if "denoise" in params["short_preparation_pipe"].keys():
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_denoised_T2',
                        brain_segment_pipe, 'inputnode.debiased_T1')

                else:
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_T2',
                        brain_segment_pipe, 'inputnode.debiased_T1')

            seg_pipe.connect(
                restore_mask_T2, 'out_file',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

    else:

        # using T1
        if "masked_correct_bias_pipe" in params.keys():
            seg_pipe.connect(
                masked_correct_bias_pipe, 'outputnode.mask_debiased_T1',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

            if "correct_bias_pipe" in params.keys():
                seg_pipe.connect(
                    correct_bias_pipe, "outputnode.debiased_T1",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "N4debias" in params.keys():
                seg_pipe.connect(
                    N4debias_T1, "output_image",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "fast" in params.keys():
                seg_pipe.connect(
                    fast_T1, ("restored_image", show_files),
                    brain_segment_pipe, 'inputnode.debiased_T1')

            else:
                if "denoise" in params["short_preparation_pipe"].keys():
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                        brain_segment_pipe, 'inputnode.debiased_T1')

                else:
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_T1',
                        brain_segment_pipe, 'inputnode.debiased_T1')

        elif "debias" in params.keys():
            seg_pipe.connect(
                debias, 't1_debiased_brain_file',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

            seg_pipe.connect(
                debias, 't1_debiased_file',
                brain_segment_pipe, 'inputnode.debiased_T1')

        else:

            if "correct_bias_pipe" in params.keys():
                seg_pipe.connect(
                    correct_bias_pipe, "outputnode.debiased_T1",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "N4debias" in params.keys():
                seg_pipe.connect(
                    N4debias_T1, "output_image",
                    brain_segment_pipe, 'inputnode.debiased_T1')

            elif "fast" in params.keys():
                seg_pipe.connect(
                    fast_T1, ("restored_image", show_files),
                    brain_segment_pipe, 'inputnode.debiased_T1')

            else:
                if "denoise" in params["short_preparation_pipe"].keys():
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                        brain_segment_pipe, 'inputnode.debiased_T1')

                else:
                    seg_pipe.connect(
                        data_preparation_pipe, 'outputnode.stereo_T1',
                        brain_segment_pipe, 'inputnode.debiased_T1')

            seg_pipe.connect(
                restore_mask_T1, 'out_file',
                brain_segment_pipe, 'inputnode.masked_debiased_T1')

    # outputnode
    seg_pipe.connect(brain_segment_pipe, 'outputnode.segmented_file',
                     outputnode, 'stereo_segmented_brain_mask')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_gm',
                     outputnode, 'stereo_prob_gm')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_wm',
                     outputnode, 'stereo_prob_wm')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_csf',
                     outputnode, 'stereo_prob_csf')

    if pad and space == "native":
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.segmented_file",
            outputnode, "native_segmented_brain_mask", params)
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_gm",
            outputnode, "native_prob_gm", params,
            inter_val="LIN")
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_wm",
            outputnode, "native_prob_wm", params,
            inter_val="LIN")
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_csf",
            outputnode, "native_prob_csf", params,
            inter_val="LIN")

    if "pad_template" in params["short_preparation_pipe"].keys():
        pad_stereo_stereo_brain_mask = NodeParams(
            ImageMath(),
            params=parse_key(params["short_preparation_pipe"],
                             "pad_template"),
            name="pad_stereo_stereo_brain_mask")

        seg_pipe.connect(
            brain_segment_pipe, "outputnode.segmented_file",
            pad_stereo_stereo_brain_mask, "op1")

        seg_pipe.connect(
            pad_stereo_stereo_brain_mask, "output_image",
            outputnode, "stereo_padded_segmented_brain_mask")

    # ############################################## export 5tt

    if "export_5tt_pipe" in params["brain_segment_pipe"]:

        seg_pipe.connect(brain_segment_pipe, 'outputnode.gen_5tt',
                         outputnode, 'stereo_gen_5tt')

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                brain_segment_pipe, "outputnode.gen_5tt",
                outputnode, "native_gen_5tt", params)

    # ############################################## surface

    if "nii2mesh_brain_pipe" in params["brain_segment_pipe"].keys():

        nii2mesh_brain_pipe = create_nii2mesh_brain_pipe(
            params=parse_key(params["brain_segment_pipe"],
                             "nii2mesh_brain_pipe"))

        seg_pipe.connect(brain_segment_pipe, "outputnode.segmented_file",
                         nii2mesh_brain_pipe, 'inputnode.segmented_file')

        seg_pipe.connect(nii2mesh_brain_pipe, "outputnode.wmgm_stl",
                         outputnode, 'wmgm_stl')

        seg_pipe.connect(nii2mesh_brain_pipe, "outputnode.wmgm_nii",
                         outputnode, 'stereo_wmgm_mask')

        if pad:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                nii2mesh_brain_pipe, "outputnode.wmgm_nii",
                outputnode, "native_wmgm_mask", params)

    elif "IsoSurface_brain_pipe" in params["brain_segment_pipe"].keys():

        IsoSurface_brain_pipe = create_IsoSurface_brain_pipe(
            params=parse_key(params["brain_segment_pipe"],
                             "IsoSurface_brain_pipe"))

        seg_pipe.connect(brain_segment_pipe, "outputnode.segmented_file",
                         IsoSurface_brain_pipe, 'inputnode.segmented_file')

        seg_pipe.connect(IsoSurface_brain_pipe, "outputnode.wmgm_stl",
                         outputnode, 'wmgm_stl')

        seg_pipe.connect(IsoSurface_brain_pipe, "outputnode.wmgm_nii",
                         outputnode, 'stereo_wmgm_mask')

        if pad:
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                IsoSurface_brain_pipe, "outputnode.wmgm_nii",
                outputnode, "native_wmgm_mask", params)

    elif 'nii_to_mesh_pipe' in params.keys():
        # kept for compatibility but nii2mesh is prefered...
        nii_to_mesh_pipe = create_nii_to_mesh_pipe(
            params_template=params_template,
            params=parse_key(params, "nii_to_mesh_pipe"))

        # from data_preparation_pipe
        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                nii_to_mesh_pipe, 'inputnode.t1_ref_file')

        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                nii_to_mesh_pipe, 'inputnode.t1_ref_file')

        # from brain_segment_pipe
        seg_pipe.connect(brain_segment_pipe,
                         'register_NMT_pipe.NMT_subject_align.warpinv_file',
                         nii_to_mesh_pipe, 'inputnode.warpinv_file')

        seg_pipe.connect(
            brain_segment_pipe,
            'register_NMT_pipe.NMT_subject_align.inv_transfo_file',
            nii_to_mesh_pipe, 'inputnode.inv_transfo_file')

        seg_pipe.connect(brain_segment_pipe,
                         'register_NMT_pipe.NMT_subject_align.aff_file',
                         nii_to_mesh_pipe, 'inputnode.aff_file')

        seg_pipe.connect(brain_segment_pipe,
                         'segment_atropos_pipe.outputnode.segmented_file',
                         nii_to_mesh_pipe, "inputnode.segmented_file")

    return seg_pipe


###############################################################################
# ANTS based segmentation for adrien baboons (T1 without T2)
# -soft ANTS_T1


def create_full_T1_ants_subpipes(params_template, params_template_stereo,
                                 params={}, name="full_T1_ants_subpipes",
                                 mask_file=None,
                                 space="native", pad=False):
    """
    Description: Full pipeline to segment T1 (with no T2).

    Params:

    - short_preparation_pipe (see :class:`create_short_preparation_pipe <\
    macapype.pipelines.prepare.create_short_preparation_pipe>`
    - brain_extraction_T1_pipe (see :class:`create_brain_extraction_T1_pipe \
    <macapype.pipelines.full_pipelines.create_brain_extraction_T1_pipe>`)
    - brain_segment_T1_pipe (see \
    :class:`create_brain_segment_from_mask_T1_pipe \
<macapype.pipelines.full_pipelines.create_brain_segment_from_mask_T1_pipe>`)

    Inputs:

        inputnode:

            list_T1:
                preprocessed T1 file name

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params_template:
                dictionary of template files

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "full_segment_pipe")

    Outputs:

    """

    # creating pipeline
    seg_pipe = pe.Workflow(name=name)

    # Creating input node (grab only T1 files)
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'indiv_params']),
        name='inputnode')

    # output node
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                    "native_T1",
                    'stereo_T1',
                    'stereo_denoised_T1',
                    "stereo_padded_T1",

                    'stereo_brain_mask', 'native_brain_mask',
                    'stereo_debiased_T1', "native_debiased_T1",

                    "native_masked_debiased_T1", "stereo_masked_debiased_T1",

                    'native_segmented_brain_mask', "native_gen_5tt",
                    'native_prob_gm', 'native_prob_wm', 'native_prob_csf',

                    'stereo_segmented_brain_mask', "stereo_gen_5tt",
                    'stereo_prob_gm', 'stereo_prob_wm', 'stereo_prob_csf',

                    "native_wmgm_mask",
                    "stereo_wmgm_mask",
                    "wmgm_stl",

                    "stereo_to_native_trans",
                    "native_to_stereo_trans"]),
        name='outputnode')

    # preprocessing (perform preparation pipe with only T1)
    if 'short_preparation_pipe' in params.keys():
        data_preparation_pipe = create_short_preparation_T1_pipe(
            params=parse_key(params, "short_preparation_pipe"),
            params_template=params_template_stereo)

    else:
        print("Error, short_preparation_pipe was not found in params, \
            skipping")
        return seg_pipe

    seg_pipe.connect(inputnode, 'list_T1',
                     data_preparation_pipe, 'inputnode.list_T1')

    seg_pipe.connect(inputnode, 'indiv_params',
                     data_preparation_pipe, 'inputnode.indiv_params')

    # outputnode
    seg_pipe.connect(data_preparation_pipe, 'outputnode.native_T1',
                     outputnode, 'native_T1')

    seg_pipe.connect(data_preparation_pipe, 'outputnode.stereo_T1',
                     outputnode, 'stereo_T1')

    if "denoise" in params["short_preparation_pipe"].keys():
        seg_pipe.connect(
            data_preparation_pipe, 'outputnode.stereo_denoised_T1',
            outputnode, 'stereo_denoised_T1')

    seg_pipe.connect(
        data_preparation_pipe, "outputnode.stereo_padded_T1",
        outputnode, "stereo_padded_T1")

    seg_pipe.connect(
        data_preparation_pipe, "outputnode.native_to_stereo_trans",
        outputnode, 'native_to_stereo_trans')

    seg_pipe.connect(
        data_preparation_pipe, "outputnode.stereo_to_native_trans",
        outputnode, 'stereo_to_native_trans')

    # ######### correct_bias
    if "N4debias" in params.keys():

        print("Found N4debias in params.json")

        # N4 intensity normalization over T1
        N4debias_T1 = NodeParams(ants.N4BiasFieldCorrection(),
                                 params=parse_key(params, "N4debias"),
                                 name='N4debias_T1')

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                N4debias_T1, "input_image")

        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                N4debias_T1, "input_image")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "N4debias"),
            N4debias_T1, "indiv_params")

        # outputnode
        seg_pipe.connect(N4debias_T1, "output_image",
                         outputnode, "stereo_debiased_T1")

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                N4debias_T1, "output_image",
                outputnode, "native_debiased_T1", params,
                inter_val="LIN")

    elif "fast" in params.keys():

        print("Found fast in params.json")

        # fast over T1
        fast_T1 = NodeParams(
            fsl.FAST(),
            params=parse_key(params, "fast"),
            name='fast_T1')

        fast_T1.inputs.output_biascorrected = True
        fast_T1.inputs.output_biasfield = True
        fast_T1.inputs.img_type = 1

        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                fast_T1, "in_files")
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                fast_T1, "in_files")

        seg_pipe.connect(
            inputnode, ('indiv_params', parse_key, "fast"),
            fast_T1, "indiv_params")

        # outputnode
        seg_pipe.connect(fast_T1, "restored_image",
                         outputnode, "stereo_debiased_T1")

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                fast_T1, "restored_image",
                outputnode, "native_debiased_T1", params,
                inter_val="LIN")
    else:
        print("No debias is applied before masking")

    #  extract brain pipeline
    if "extract_pipe" not in params.keys():
        print("Error, extract_pipe was not found in params, \
            skipping")
        return seg_pipe

    if mask_file is None:

        extract_T1_pipe = create_extract_pipe(
            params_template=params_template,
            params=parse_key(params, "extract_pipe"))

        seg_pipe.connect(
            inputnode, "indiv_params",
            extract_T1_pipe, "inputnode.indiv_params")

        if "N4debias" in params.keys():

            # brain extraction
            seg_pipe.connect(N4debias_T1, "output_image",
                             extract_T1_pipe, "inputnode.restore_T1")

        elif "fast" in params.keys():

            # brain extraction
            seg_pipe.connect(
                fast_T1, "restored_image",
                extract_T1_pipe, "inputnode.restore_T1")

        else:
            # brain extraction (with atlasbrex)
            if "denoise" in params["short_preparation_pipe"].keys():
                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                    extract_T1_pipe, "inputnode.restore_T1")
            else:
                seg_pipe.connect(
                    data_preparation_pipe, 'outputnode.stereo_T1',
                    extract_T1_pipe, "inputnode.restore_T1")

        # outputnode
        seg_pipe.connect(
            extract_T1_pipe, "smooth_mask.out_file",
            outputnode, "stereo_brain_mask")

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                extract_T1_pipe, "smooth_mask.out_file",
                outputnode, "native_brain_mask", params)

    else:
        print("Using native external mask {}".format(mask_file))
        outputnode.inputs.native_brain_mask = mask_file

        # apply transfo to list
        apply_crop_external_mask = pe.Node(RegResample(inter_val="NN"),
                                           name='apply_crop_external_mask')

        apply_crop_external_mask.inputs.flo_file = mask_file

        seg_pipe.connect(data_preparation_pipe,
                         'outputnode.native_to_stereo_trans',
                         apply_crop_external_mask, "trans_file")

        seg_pipe.connect(data_preparation_pipe, "outputnode.stereo_T1",
                         apply_crop_external_mask, "ref_file")

        # outputnode
        seg_pipe.connect(apply_crop_external_mask, "out_file",
                         outputnode, "stereo_brain_mask")

    # restore_mask_T1
    restore_mask_T1 = pe.Node(fsl.ApplyMask(), name='restore_mask_T1')

    if "N4debias" in params.keys():

        # brain extraction
        seg_pipe.connect(N4debias_T1, "output_image",
                         restore_mask_T1, 'in_file')

    elif "fast" in params.keys():

        # brain extraction
        seg_pipe.connect(
            fast_T1, ("restored_image", show_files),
            restore_mask_T1, 'in_file')

    else:
        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                restore_mask_T1, 'in_file')
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                restore_mask_T1, 'in_file')

    if mask_file is None:
        seg_pipe.connect(
            extract_T1_pipe, "smooth_mask.out_file",
            restore_mask_T1, 'mask_file')

    else:
        seg_pipe.connect(
            apply_crop_external_mask, "out_file",
            restore_mask_T1, 'mask_file')

    seg_pipe.connect(restore_mask_T1, 'out_file',
                     outputnode, 'stereo_masked_debiased_T1')

    if pad and space == "native":
        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            restore_mask_T1, 'out_file',
            outputnode, "native_masked_debiased_T1", params,
            inter_val="LIN")

    # ### full_segment (restarting from the avg_align files)
    if "brain_segment_pipe" not in params.keys():
        print("Error, brain_segment_pipe was not found in params, \
            skipping")
        return seg_pipe

    brain_segment_pipe = create_brain_segment_from_mask_pipe(
        params_template=params_template,
        params=parse_key(params, "brain_segment_pipe"), space=space)

    if "N4debias" in params.keys():
        # brain extraction
        seg_pipe.connect(N4debias_T1, "output_image",
                         brain_segment_pipe, 'inputnode.debiased_T1')

    elif "fast" in params.keys():
        # brain extraction
        seg_pipe.connect(
            fast_T1, ("restored_image", show_files),
            brain_segment_pipe, 'inputnode.debiased_T1')

    else:
        if "denoise" in params["short_preparation_pipe"].keys():
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_denoised_T1',
                brain_segment_pipe, 'inputnode.debiased_T1')
        else:
            seg_pipe.connect(
                data_preparation_pipe, 'outputnode.stereo_T1',
                brain_segment_pipe, 'inputnode.debiased_T1')

    seg_pipe.connect(restore_mask_T1, 'out_file',
                     brain_segment_pipe, 'inputnode.masked_debiased_T1')

    seg_pipe.connect(inputnode, 'indiv_params',
                     brain_segment_pipe, 'inputnode.indiv_params')

    # outputnode
    seg_pipe.connect(brain_segment_pipe, 'outputnode.segmented_file',
                     outputnode, 'stereo_segmented_brain_mask')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_csf',
                     outputnode, 'stereo_prob_csf')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_gm',
                     outputnode, 'stereo_prob_gm')

    seg_pipe.connect(brain_segment_pipe, 'outputnode.prob_wm',
                     outputnode, 'stereo_prob_wm')

    if pad and space == "native":

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.segmented_file",
            outputnode, "native_segmented_brain_mask", params)

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_gm",
            outputnode, "native_prob_gm", params,
            inter_val="LIN")

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_wm",
            outputnode, "native_prob_wm", params,
            inter_val="LIN")

        pad_back(
            seg_pipe, data_preparation_pipe, inputnode,
            brain_segment_pipe, "outputnode.prob_csf",
            outputnode, "native_prob_csf", params,
            inter_val="LIN")

    if "export_5tt_pipe" in params["brain_segment_pipe"]:

        seg_pipe.connect(brain_segment_pipe, 'outputnode.gen_5tt',
                         outputnode, 'stereo_gen_5tt')
        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                brain_segment_pipe, "outputnode.gen_5tt",
                outputnode, "native_gen_5tt", params)

    if "nii2mesh_brain_pipe" in params["brain_segment_pipe"]:

        nii2mesh_brain_pipe = create_nii2mesh_brain_pipe(
            params=parse_key(params["brain_segment_pipe"],
                             "nii2mesh_brain_pipe"))

        seg_pipe.connect(brain_segment_pipe, "outputnode.segmented_file",
                         nii2mesh_brain_pipe, 'inputnode.segmented_file')

        # outputnode
        seg_pipe.connect(nii2mesh_brain_pipe, "outputnode.wmgm_stl",
                         outputnode, 'wmgm_stl')

        seg_pipe.connect(nii2mesh_brain_pipe, "outputnode.wmgm_nii",
                         outputnode, 'stereo_wmgm_mask')

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                nii2mesh_brain_pipe, "outputnode.wmgm_nii",
                outputnode, "native_wmgm_mask", params)

    elif "IsoSurface_brain_pipe" in params["brain_segment_pipe"].keys():

        IsoSurface_brain_pipe = create_IsoSurface_brain_pipe(
            params=parse_key(params["brain_segment_pipe"],
                             "IsoSurface_brain_pipe"))

        seg_pipe.connect(brain_segment_pipe, "outputnode.segmented_file",
                         IsoSurface_brain_pipe, 'inputnode.segmented_file')

        # outputnode
        seg_pipe.connect(IsoSurface_brain_pipe, "outputnode.wmgm_stl",
                         outputnode, 'wmgm_stl')

        seg_pipe.connect(IsoSurface_brain_pipe, "outputnode.wmgm_nii",
                         outputnode, 'stereo_wmgm_mask')

        if pad and space == "native":
            pad_back(
                seg_pipe, data_preparation_pipe, inputnode,
                IsoSurface_brain_pipe, "outputnode.wmgm_nii",
                outputnode, "native_wmgm_mask", params)

    return seg_pipe




import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe

import nipype.interfaces.fsl as fsl

from nipype.interfaces.niftyreg import reg, regutils

from nipype.interfaces.ants.utils import ImageMath

from nipype.interfaces.ants.segmentation import DenoiseImage

from ..utils.utils_nodes import NodeParams
from ..utils.misc import parse_key

from ..nodes.prepare import average_align

from ..nodes.register import pad_zero_mri

# should be in nipype code directly
from ..nodes.prepare import Refit

from .register import create_crop_aladin_pipe


def _create_avg_reorient_pipeline(name="avg_reorient_pipe", params={}):
    """
    By david:
    average_align
    3drefit -deoblique -origin LSP image_good.nii.gz;
    fslreorient2std image_good.nii.gz
    """

    # creating pipeline
    reorient_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_img', 'indiv_params']),
        name='inputnode'
    )

    average_img = pe.Node(
        niu.Function(input_names=['list_img'],
                     output_names=['avg_img'],
                     function=average_align),
        name="average_img")

    reorient_pipe.connect(inputnode, 'list_img', average_img, 'list_img')

    # Refit
    reorient = NodeParams(Refit(),
                          params=parse_key(params, "reorient"),
                          name="reorient")

    reorient_pipe.connect(average_img, 'avg_img', reorient, 'in_file')

    reorient_pipe.connect(inputnode, ('indiv_params', parse_key, 'reorient'),
                          reorient, 'indiv_params')

    # Reorient2std
    std_img = pe.Node(fsl.Reorient2Std(), name="std_img")

    reorient_pipe.connect(reorient, "out_file", std_img, 'in_file')

    outputnode = pe.Node(niu.IdentityInterface(fields=['avg_img', 'std_img']),
                         name="outputnode")

    reorient_pipe.connect(std_img, 'out_file', outputnode, 'std_img')

    return reorient_pipe


###############################################################################
# main pipeline: "short_preparation_pipe"
###############################################################################


def create_short_preparation_pipe(params, params_template={},
                                  name="short_preparation_pipe"):
    """Description: short data preparation (average, reorient, crop/betcrop \
    and denoise)

    Processing steps;

    - T1s and T2s are averaged (by modality).
    - reorient (opt) makes the whole pipeline more neat)
    - Cropped (bet_crop include alignement, and mask generation)
    - denoising all images (no denoise_first)

    Params:

    - reorient (optional; new_dims as a string (e.g. "x z -y"))
    - crop (see `ExtractRoi <https://nipype.readthedocs.io/en/0.12.1/inte\
    rfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ for \
    arguments). crop is also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://\
    nipype.readthedocs.io/en/0.12.1/interfaces/generated/nipype.interfaces\
    .ants.segmentation.html#denoiseimage>`_ for arguments))

    Inputs:

        inputnode:

            list_T1:
                T1 files (from BIDSDataGrabber)

            list_T2:
                T2 files

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "long_multi_preparation_pipe")

    Outputs:

        outputnode:

            preproc_T1:
                preprocessed T1 file

            preproc_T2:
                preprocessed T2 file
    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'list_T2', 'indiv_params']),
        name='inputnode'
    )

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['native_T1', 'native_T2',
                                      'stereo_T1', 'stereo_T2',
                                      'stereo_denoised_T1',
                                      'stereo_denoised_T2',
                                      "stereo_padded_T1",
                                      "stereo_padded_T2",
                                      "native_to_stereo_trans",
                                      "stereo_to_native_trans"]),
        name='outputnode')

    if "avg_reorient_pipe" in params.keys():

        print("Found avg_reorient_pipe for av_T1 and av_T2")

        av_T1 = _create_avg_reorient_pipeline(
            name="av_T1", params=parse_key(params, "avg_reorient_pipe"))

        if "use_T2" in params.keys():

            data_preparation_pipe.connect(
                inputnode, 'list_T2',
                av_T1, 'inputnode.list_img')
        else:

            data_preparation_pipe.connect(
                inputnode, 'list_T1',
                av_T1, 'inputnode.list_img')

        data_preparation_pipe.connect(inputnode, 'indiv_params',
                                      av_T1, 'inputnode.indiv_params')

        av_T2 = _create_avg_reorient_pipeline(
            name="av_T2", params=parse_key(params, "avg_reorient_pipe"))

        if "use_T2" in params.keys():

            data_preparation_pipe.connect(
                inputnode, 'list_T1',
                av_T2, 'inputnode.list_img')
        else:

            data_preparation_pipe.connect(
                inputnode, 'list_T2',
                av_T2, 'inputnode.list_img')

        data_preparation_pipe.connect(inputnode, 'indiv_params',
                                      av_T2, 'inputnode.indiv_params')

    else:

        # avererge if multiple T1
        av_T1 = pe.Node(
            niu.Function(input_names=['list_img'],
                         output_names=['avg_img'],
                         function=average_align),
            name="av_T1")

        # average if multiple T2
        av_T2 = pe.Node(niu.Function(
            input_names=['list_img'],
            output_names=['avg_img'],
            function=average_align), name="av_T2")

        if "use_T2" in params.keys():

            data_preparation_pipe.connect(
                inputnode, 'list_T1',
                av_T2, 'list_img')

            data_preparation_pipe.connect(
                inputnode, 'list_T2',
                av_T1, 'list_img')
        else:

            data_preparation_pipe.connect(
                inputnode, 'list_T1',
                av_T1, 'list_img')

            data_preparation_pipe.connect(
                inputnode, 'list_T2',
                av_T2, 'list_img')

    # align avg T2 on avg T1
    if 'aladin_T2_on_T1' in params.keys():

        align_T2_on_T1 = pe.Node(reg.RegAladin(), name="align_T2_on_T1")
        align_T2_on_T1.inputs.rig_only_flag = True

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(av_T1, 'outputnode.std_img',
                                          align_T2_on_T1, 'ref_file')

            data_preparation_pipe.connect(av_T2, 'outputnode.std_img',
                                          align_T2_on_T1, 'flo_file')

        else:
            data_preparation_pipe.connect(av_T1, 'avg_img',
                                          align_T2_on_T1, 'ref_file')

            data_preparation_pipe.connect(av_T2, 'avg_img',
                                          align_T2_on_T1, 'flo_file')

        # pad image T2
        pad_image_T2 = pe.Node(
            ImageMath(),
            name="pad_image_T2")

        pad_image_T2.inputs.copy_header = True
        pad_image_T2.inputs.operation = "PadImage"
        pad_image_T2.inputs.op2 = '200'

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T2, 'outputnode.std_img',
                pad_image_T2, "op1")

        else:
            data_preparation_pipe.connect(
                av_T2, 'avg_img',
                pad_image_T2, "op1")

        # resampling using transfo on much bigger image
        reg_resample_T2 = pe.Node(
            regutils.RegResample(pad_val=0.0),
            name="reg_resample_T2")

        data_preparation_pipe.connect(
            align_T2_on_T1, 'aff_file',
            reg_resample_T2, 'trans_file')

        data_preparation_pipe.connect(
            pad_image_T2, 'output_image',
            reg_resample_T2, "flo_file")

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                reg_resample_T2, "ref_file")

        else:
            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                reg_resample_T2, "ref_file")

    # align avg T2 on avg T1
    else:

        align_T2_on_T1 = pe.Node(fsl.FLIRT(), name="align_T2_on_T1")
        align_T2_on_T1.inputs.dof = 6

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(av_T1, 'outputnode.std_img',
                                          align_T2_on_T1, 'reference')

            data_preparation_pipe.connect(av_T2, 'outputnode.std_img',
                                          align_T2_on_T1, 'in_file')

        else:
            data_preparation_pipe.connect(av_T1, 'avg_img',
                                          align_T2_on_T1, 'reference')

            data_preparation_pipe.connect(av_T2, 'avg_img',
                                          align_T2_on_T1, 'in_file')

    # auto cropping and register to stereo
    if "crop_T1" in params.keys():
        print('crop_T1 is in params')

        assert "args" in params["crop_T1"].keys(), \
            "Error, args is not specified for crop node, breaking"

        # cropping
        # Crop bounding box for T1
        crop_T1 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop_T1'),
                             name='crop_T1')

        # Crop bounding box for T2
        crop_T2 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop_T1'),
                             name='crop_T2')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T1, 'indiv_params')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T2, 'indiv_params')

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(av_T1, 'outputnode.std_img',
                                          crop_T1, 'in_file')
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                outputnode, 'native_T1')

        else:
            data_preparation_pipe.connect(av_T1, 'avg_img',
                                          crop_T1, 'in_file')

            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                outputnode, 'native_T1')

        if 'aladin_T2_on_T1' in params.keys():
            data_preparation_pipe.connect(
                reg_resample_T2, 'out_file',
                crop_T2, 'in_file')

            data_preparation_pipe.connect(
                reg_resample_T2, 'out_file',
                outputnode, 'native_T2')

        else:
            data_preparation_pipe.connect(
                align_T2_on_T1, "out_file",
                crop_T2, 'in_file')

            data_preparation_pipe.connect(
                align_T2_on_T1, "out_file",
                outputnode, 'native_T2')

    else:

        # outputnode
        if "use_T2" in params.keys():

            if "avg_reorient_pipe" in params.keys():
                data_preparation_pipe.connect(
                    av_T1, 'outputnode.std_img',
                    outputnode, 'native_T2')
            else:
                data_preparation_pipe.connect(
                    av_T1, 'avg_img',
                    outputnode, 'native_T2')

            if 'aladin_T2_on_T1' in params.keys():
                data_preparation_pipe.connect(
                    reg_resample_T2, 'out_file',
                    outputnode, 'native_T1')
            else:
                data_preparation_pipe.connect(
                    align_T2_on_T1, "out_file",
                    outputnode, 'native_T1')

        else:

            if "avg_reorient_pipe" in params.keys():
                data_preparation_pipe.connect(
                    av_T1, 'outputnode.std_img',
                    outputnode, 'native_T1')
            else:
                data_preparation_pipe.connect(
                    av_T1, 'avg_img',
                    outputnode, 'native_T1')

            if 'aladin_T2_on_T1' in params.keys():
                data_preparation_pipe.connect(
                    reg_resample_T2, 'out_file',
                    outputnode, 'native_T2')
            else:
                data_preparation_pipe.connect(
                    align_T2_on_T1, "out_file",
                    outputnode, 'native_T2')

    # register T1 to stereo image
    crop_aladin_pipe = create_crop_aladin_pipe(
        "crop_aladin_pipe",
        params=parse_key(params, "crop_aladin_pipe"))

    if "crop_T1" in params.keys():
        data_preparation_pipe.connect(
            crop_T1, "roi_file",
            crop_aladin_pipe, 'inputnode.native_T1')

    else:
        # connect orig_native_T1
        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                crop_aladin_pipe, 'inputnode.native_T1')

        else:
            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                crop_aladin_pipe, 'inputnode.native_T1')

    data_preparation_pipe.connect(
        inputnode, 'indiv_params',
        crop_aladin_pipe, 'inputnode.indiv_params')

    crop_aladin_pipe.inputs.inputnode.stereo_template_T1 = \
        params_template["template_head"]

    # apply reg_resample to T2
    apply_crop_aladin_T2 = NodeParams(
        regutils.RegResample(),
        name='apply_crop_aladin_T2')

    if 'aladin_T2_on_T1' in params.keys():
        data_preparation_pipe.connect(
            reg_resample_T2, 'out_file',
            apply_crop_aladin_T2, 'flo_file')

    else:
        data_preparation_pipe.connect(align_T2_on_T1, "out_file",
                                      apply_crop_aladin_T2, 'flo_file')

    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
        apply_crop_aladin_T2, 'trans_file')

    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.stereo_T1',
        apply_crop_aladin_T2, 'ref_file')

    # compute inv transfo
    inv_tranfo = NodeParams(
        regutils.RegTransform(),
        params=parse_key(params, "inv_transfo_aladin"),
        name='inv_tranfo')

    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
        inv_tranfo, 'inv_aff_input')

    # outputnode (transfo)
    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
        outputnode, 'native_to_stereo_trans')

    data_preparation_pipe.connect(
        inv_tranfo, 'out_file',
        outputnode, 'stereo_to_native_trans')

    # outputnode (stereo)
    if "use_T2" in params.keys():
        data_preparation_pipe.connect(
            crop_aladin_pipe, "outputnode.stereo_T1",
            outputnode, 'stereo_T2')

        data_preparation_pipe.connect(
            apply_crop_aladin_T2, 'out_file',
            outputnode, 'stereo_T1')

    else:

        data_preparation_pipe.connect(
            crop_aladin_pipe, "outputnode.stereo_T1",
            outputnode, 'stereo_T1')

        data_preparation_pipe.connect(
            apply_crop_aladin_T2, 'out_file',
            outputnode, 'stereo_T2')

    # denoise with Ants package
    if "denoise" in params.keys():

        denoise_T1 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T1")

        denoise_T2 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T2")
        # inputs
        data_preparation_pipe.connect(
            crop_aladin_pipe, "outputnode.stereo_T1",
            denoise_T1, 'input_image')

        data_preparation_pipe.connect(
            apply_crop_aladin_T2, 'out_file',
            denoise_T2, 'input_image')

        # outputs
        if "use_T2" in params.keys():

            data_preparation_pipe.connect(
                denoise_T1, 'output_image',
                outputnode, 'stereo_denoised_T2')

            data_preparation_pipe.connect(
                denoise_T2, 'output_image',
                outputnode, 'stereo_denoised_T1')
        else:

            data_preparation_pipe.connect(
                denoise_T1, 'output_image',
                outputnode, 'stereo_denoised_T1')

            data_preparation_pipe.connect(
                denoise_T2, 'output_image',
                outputnode, 'stereo_denoised_T2')

    # resample T1 to higher dimension
    if "pad_template" in params.keys():

        resample_T1_pad = pe.Node(
            regutils.RegResample(),
            name="resample_T1_pad")

        resample_T2_pad = pe.Node(
            regutils.RegResample(),
            name="resample_T2_pad")

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                resample_T1_pad, "flo_file")
        else:
            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                resample_T1_pad, "flo_file")

        if 'aladin_T2_on_T1' in params.keys():
            data_preparation_pipe.connect(
                reg_resample_T2, 'out_file',
                resample_T2_pad, "flo_file")
        else:
            data_preparation_pipe.connect(
                align_T2_on_T1, "out_file",
                outputnode, 'native_T2')

        if "padded_template_head" in params_template.keys():
            print("Found padded_template_head in template_stereo")

            resample_T1_pad.inputs.ref_file = \
                params_template["padded_template_head"]

            resample_T2_pad.inputs.ref_file = \
                params_template["padded_template_head"]

        elif "template_head" in params_template.keys():

            print("***** Found pad_template and using template_head ")
            pad_template = NodeParams(
                ImageMath(),
                params=parse_key(params, "pad_template"),
                name="pad_template")

            pad_template.inputs.op1 = params_template["template_head"]

            #  resample_T1_pad
            data_preparation_pipe.connect(
                pad_template, 'output_image',
                resample_T1_pad, "ref_file")

            #  resample_T2_pad
            data_preparation_pipe.connect(
                pad_template, 'output_image',
                resample_T2_pad, "ref_file")

        else:
            print("Error, template_head or padded_template_head should be \
                defined in template")
            exit(-1)

        data_preparation_pipe.connect(
            crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
            resample_T1_pad, "trans_file")

        data_preparation_pipe.connect(
            crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
            resample_T2_pad, "trans_file")

        # outputnode
        if "use_T2" in params.keys():

            data_preparation_pipe.connect(
                resample_T1_pad, 'out_file',
                outputnode, 'stereo_padded_T2')

            data_preparation_pipe.connect(
                resample_T2_pad, 'out_file',
                outputnode, 'stereo_padded_T1')

        else:

            data_preparation_pipe.connect(
                resample_T1_pad, 'out_file',
                outputnode, 'stereo_padded_T1')

            data_preparation_pipe.connect(
                resample_T2_pad, 'out_file',
                outputnode, 'stereo_padded_T2')

    return data_preparation_pipe


###############################################################################
# works with only one T1
def create_short_preparation_T1_pipe(params, params_template,
                                     name="short_preparation_T1_pipe"):
    """Description: T1 only short preparation

    Processing steps:

    - T1s and T2s are averaged (by modality).
    - reorient (opt: reorient) makes the whole pipeline more neat)
    - Cropped (bet_crop include alignement, and mask generation)
    - denoising all images (no denoise_first)

    Params:

    - reorient (optional; new_dims as a string (e.g. "x z -y"))
    - bet_crop (:class:`T1xT2BET <macapype.nodes.extract_brain.T1xT2BET>`\
    ) or crop (see `ExtractRoi <https://nipype.readthedocs.io/en/0.12.1/inte\
    rfaces/generated/nipype.interfaces.fsl.utils.html#extractroi>`_ for \
    arguments). crop is also available as :ref:`indiv_params <indiv_params>`
    - denoise (see `DenoiseImage <https://\
    nipype.readthedocs.io/en/0.12.1/interfaces/generated/nipype.interfaces\
    .ants.segmentation.html#denoiseimage>`_ for arguments))

    Inputs:

        inputnode:

            list_T1:
                T1 files (from BIDSDataGrabber)

            indiv_params (opt):
                dict with individuals parameters for some nodes

        arguments:

            params:
                dictionary of node sub-parameters (from a json file)

            name:
                pipeline name (default = "short_preparation_T1_pipe")

    Outputs:

        outputnode:

            preproc_T1:
                preprocessed T1 file

    """

    # creating pipeline
    data_preparation_pipe = pe.Workflow(name=name)

    # Creating input node
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['list_T1', 'indiv_params']),
        name='inputnode'
    )

    # Creating output node
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['stereo_T1', 'native_T1',
                                      'stereo_padded_T1',
                                      'stereo_denoised_T1',
                                      "stereo_to_native_trans",
                                      "native_to_stereo_trans"]),
        name='outputnode')

    # average if multiple T1
    if "avg_reorient_pipe" in params.keys():

        av_T1 = _create_avg_reorient_pipeline(
            "av_T1", params=parse_key(params, "avg_reorient_pipe"))

        data_preparation_pipe.connect(inputnode, 'list_T1',
                                      av_T1, 'inputnode.list_img')

        data_preparation_pipe.connect(inputnode, 'indiv_params',
                                      av_T1, 'inputnode.indiv_params')
    else:
        av_T1 = pe.Node(
            niu.Function(input_names=['list_img'],
                         output_names=['avg_img'],
                         function=average_align),
            name="av_T1")
        data_preparation_pipe.connect(inputnode, 'list_T1', av_T1, 'list_img')

    # outputnode
    if "avg_reorient_pipe" in params.keys():
        data_preparation_pipe.connect(
            av_T1, 'outputnode.std_img',
            outputnode, 'native_T1')

    else:
        data_preparation_pipe.connect(av_T1, 'avg_img',
                                      outputnode, 'native_T1')

    # cropping and possibly sending to stereo space
    if "crop_T1" in params.keys():
        print('crop_T1 is in params')

        assert "args" in params["crop_T1"].keys(), \
            "Error, args is not specified for crop node, breaking"

        # cropping
        # Crop bounding box for T1
        crop_T1 = NodeParams(fsl.ExtractROI(),
                             params=parse_key(params, 'crop'),
                             name='crop_T1')

        if "avg_reorient_pipe" in params.keys():

            data_preparation_pipe.connect(av_T1,  'outputnode.std_img',
                                          crop_T1, 'in_file')
        else:

            data_preparation_pipe.connect(av_T1,  'avg_img',
                                          crop_T1, 'in_file')

        data_preparation_pipe.connect(
            inputnode, ("indiv_params", parse_key, "crop_T1"),
            crop_T1, 'indiv_params')

    # register T1 to stereo image
    crop_aladin_pipe = create_crop_aladin_pipe(
        "crop_aladin_pipe",
        params=parse_key(params, "crop_aladin_pipe"))

    # register T1 to stereo image
    crop_aladin_pipe = create_crop_aladin_pipe(
        "crop_aladin_pipe",
        params=parse_key(params, "crop_aladin_pipe"))

    if "crop_T1" in params.keys():
        data_preparation_pipe.connect(
            crop_T1, "roi_file",
            crop_aladin_pipe, 'inputnode.native_T1')

    else:
        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                crop_aladin_pipe, 'inputnode.native_T1')

        else:
            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                crop_aladin_pipe, 'inputnode.native_T1')

    data_preparation_pipe.connect(
        inputnode, 'indiv_params',
        crop_aladin_pipe, 'inputnode.indiv_params')

    crop_aladin_pipe.inputs.inputnode.stereo_template_T1 = \
        params_template["template_head"]

    # compute inv transfo
    inv_tranfo = NodeParams(
        regutils.RegTransform(),
        params=parse_key(params, "inv_transfo_aladin"),
        name='inv_tranfo')

    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
        inv_tranfo, 'inv_aff_input')

    # outputnode
    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
        outputnode, 'native_to_stereo_trans')

    data_preparation_pipe.connect(
        inv_tranfo, 'out_file',
        outputnode, 'stereo_to_native_trans')

    data_preparation_pipe.connect(
        crop_aladin_pipe, 'outputnode.stereo_T1',
        outputnode, 'stereo_T1')

    if "denoise" in params.keys():

        # denoise with Ants package
        denoise_T1 = NodeParams(interface=DenoiseImage(),
                                params=parse_key(params, "denoise"),
                                name="denoise_T1")

        # inputs
        if "crop_T1" in params.keys():
            data_preparation_pipe.connect(crop_T1, "roi_file",
                                          denoise_T1, 'input_image')

        else:
            data_preparation_pipe.connect(
                crop_aladin_pipe, 'outputnode.stereo_T1',
                denoise_T1, 'input_image')

        # outputs
        data_preparation_pipe.connect(denoise_T1, 'output_image',
                                      outputnode, 'stereo_denoised_T1')

    # resample T1 to higher dimension
    if "pad_template" in params.keys():

        resample_T1_pad = pe.Node(
            regutils.RegResample(),
            name="resample_T1")

        if "avg_reorient_pipe" in params.keys():
            data_preparation_pipe.connect(
                av_T1, 'outputnode.std_img',
                resample_T1_pad, "flo_file")

        else:
            data_preparation_pipe.connect(
                av_T1, 'avg_img',
                resample_T1_pad, "flo_file")

        if "padded_template_head" in params_template.keys():
            resample_T1_pad.inputs.ref_file = \
                params_template["padded_template_head"]

        elif "template_head" in params_template.keys():
            # padding versio of the template
            pad_template = NodeParams(
                niu.Function(
                    input_names=["img_file", "pad_val", "const"],
                    output_names=["padded_img_file"],
                    function=pad_zero_mri),
                params=parse_key(params, "resample_T1_pad"),
                name="pad_template")

            pad_template.inputs.img_file = params_template["template_head"]

            #  resample_T1_pad
            data_preparation_pipe.connect(
                pad_template, 'padded_img_file',
                resample_T1_pad, "ref_file")

        else:
            print("Error, template_head or padded_template_head should be \
                defined in template")
            exit(-1)

        data_preparation_pipe.connect(
            crop_aladin_pipe, 'outputnode.native_to_stereo_trans',
            resample_T1_pad, "trans_file")

        # outputnode
        data_preparation_pipe.connect(
            resample_T1_pad, 'out_file',
            outputnode, 'stereo_padded_T1')

    return data_preparation_pipe



import os.path as op

import json
import pprint


def update_preparation_params(ssoft=[], subjects=None, sessions=None,
                              params=None, indiv_params=None,
                              extra_wf_name=""):

    if "short_preparation_pipe" not in params.keys():

        print("short_preparation_pipe not found in params, \
            not modifying preparation pipe")

    else:

        count_all_sessions = 0
        count_T1_crops = 0
        count_long_crops = 0
        count_multi_long_crops = 0

        if subjects is None:
            print("For whole BIDS dir, \
                unable to assess if the indiv_params is correct")
            print("Running by default short_preparation_pipe and crop_T1")

        else:

            print("Will modify params if necessary, \
                given specified subjects and sessions;\n")

            for sub in indiv_params.keys():

                if sub.split('-')[1] not in subjects:
                    print('could not find subject {} in {}'.format(
                        sub.split('-')[1], subjects))
                    continue

                if all([key.split('-')[0] == "ses"
                        for key in indiv_params[sub].keys()]):

                    for ses in indiv_params[sub].keys():

                        if ses.split('-')[1] not in sessions:

                            print('could not find session {} in {}'.format(
                                ses.split('-')[1], sessions))
                            continue

                        count_all_sessions += 1

                        indiv = indiv_params[sub][ses]

                        print(indiv.keys())

                        if "crop_T1" in indiv.keys():
                            count_T1_crops += 1

                            if "crop_T2" in indiv.keys() and 't1' not in ssoft:

                                count_long_crops += 1

                                crop_T1_args = indiv["crop_T1"]["args"]
                                crop_T2_args = indiv["crop_T2"]["args"]
                                if isinstance(crop_T1_args, list) and \
                                        isinstance(crop_T2_args, list):

                                    count_multi_long_crops += 1

                else:
                    count_all_sessions += 1

                    indiv = indiv_params[sub]

                    print(indiv.keys())

                    if "crop_T1" in indiv.keys():
                        count_T1_crops += 1

            print("count_all_sessions {}".format(count_all_sessions))
            print("count_T1_crops {}".format(count_T1_crops))

            if count_all_sessions:
                if count_T1_crops == count_all_sessions:

                    print("**** Found crop for T1 for all \
                        sub/ses in indiv \
                        -> keeping short_preparation_pipe")

                    extra_wf_name += "_indiv_crop"

                    print("Adding crop_T1")
                    params["short_preparation_pipe"]["crop_T1"] = \
                        {"args": "should be defined in indiv"}

                else:
                    print("**** not all sub/ses have T1 crops,\
                        using autocrop ")

        return params, indiv_params, extra_wf_name


def update_params(ssoft=[], subjects=None, sessions=None,
                  params_file=None, indiv_params_file=None):

    # params
    assert op.exists(params_file), "Error with file {}".format(
        params_file)

    params = json.load(open(params_file))

    extra_wf_name = ""
    indiv_params = {}

    # indiv_params
    if indiv_params_file is None:

        print("No indiv params where found, modifing pipeline to default")

        if "short_preparation_pipe" in params.keys():

            if "crop_T1" in params["short_preparation_pipe"].keys():
                print("Deleting crop_T1")
                del params["short_preparation_pipe"]["crop_T1"]

        else:
            print('**** params modification not implemented \
                if no indiv_params ****')

    else:

        indiv_params = json.load(open(indiv_params_file))

        assert op.exists(indiv_params_file), "Error with file {}".format(
            indiv_params_file)

        pprint.pprint(indiv_params)

        params, indiv_params, extra_wf_name = update_preparation_params(
            ssoft, subjects=subjects, sessions=sessions,
            indiv_params=indiv_params, params=params, extra_wf_name="")

    # modifying crop_aladin
    params_spp = params["short_preparation_pipe"]

    if "crop_aladin_pipe" in params_spp.keys():

        params_spp_cap = params_spp["crop_aladin_pipe"]

        if "robustreg" in ssoft:
            print("Using robustreg option")
            params_spp_cap["reg_T1_on_template"] = {
                    "nac_flag": True,
                    "rig_only_flag": True,
                    "nosym_flag": True,
                    "ln_val": 12,
                    "lp_val": 10,
                    "smoo_r_val": 1.0
                    }

            params_spp_cap["reg_T1_on_template2"] = {
                    "rig_only_flag": True,
                    "nosym_flag": True,
                    "ln_val": 17,
                    "lp_val": 15,
                    "smoo_r_val": 1.0
                    }

        elif "halfrobustreg" in ssoft:
            print("Using halfrobustreg option")
            params_spp_cap["reg_T1_on_template"] = {
                    "nac_flag": True,
                    "rig_only_flag": True,
                    "nosym_flag": True,
                    "ln_val": 12,
                    "lp_val": 10,
                    "smoo_r_val": 1.0
                    }

    print("New params after modification")
    pprint.pprint(params)

    extra_wf_name += "_crop_aladin"

    # ANTS
    if "ants" in ssoft:
        print("Found ants in soft")

        if "noseg" in ssoft:
            print("Found noseg in soft")
            if "brain_segment_pipe" in params.keys():
                del params["brain_segment_pipe"]
                print("Deleting brain_segment_pipe")

        if "prep" in ssoft:
            print("Found prep in soft")

            # debias
            if "fast" in params.keys():
                del params["fast"]
                print("Deleting fast")

            if "N4debias" in params.keys():
                del params["N4debias"]
                print("Deleting N4debias")

            # brain mask
            if "extract_pipe" in params.keys():
                del params["extract_pipe"]
                print("Deleting extract_pipe")

            # masked debias
            if "masked_correct_bias_pipe" in params.keys():
                del params["masked_correct_bias_pipe"]
                print("Deleting masked_correct_bias_pipe")

            if "debias" in params.keys():
                del params["debias"]
                print("Deleting debias")

            # segment
            if "brain_segment_pipe" in params.keys():
                del params["brain_segment_pipe"]
                print("Deleting brain_segment_pipe")

    # SPM
    if "spm" in ssoft:
        print("Found spm in soft")

        if "noseg" in ssoft:
            print("Found noseg in soft")
            if "old_segment_pipe" in params.keys():
                del params["old_segment_pipe"]
                print("Deleting old_segment_pipe")

            if "mask_from_seg_pipe" in params.keys():
                del params["mask_from_seg_pipe"]
                print("Deleting mask_from_seg_pipe")

        if "prep" in ssoft:
            print("Found prep in soft")

            if "debias" in params.keys():
                del params["debias"]
                print("Deleting debias")

            if "reg" in params.keys():
                del params["reg"]
                print("Deleting debias")

            if "old_segment_pipe" in params.keys():
                del params["old_segment_pipe"]
                print("Deleting old_segment_pipe")

            if "mask_from_seg_pipe" in params.keys():
                del params["mask_from_seg_pipe"]
                print("Deleting mask_from_seg_pipe")

    print("After modif, running with params:")
    pprint.pprint(params)

    return params, indiv_params, extra_wf_name
